# This file is part of craft-platforms.
#
# Copyright 2025 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for build functions."""

from unittest.mock import Mock, call

from craft_platforms import _build


def test_get_snapcraft_build_plan(monkeypatch):
    fake_build_plan = Mock()
    monkeypatch.setitem(_build._APP_SPECIFIC_PLANNERS, "snapcraft", fake_build_plan)

    project_data = {
        "base": "core22",
        "build-base": "core24",
        "platforms": {},
        "type": "base",
    }
    _build.get_build_plan("snapcraft", project_data=project_data)

    assert fake_build_plan.mock_calls == [
        call(base="core22", build_base="core24", platforms={}, snap_type="base")
    ]



# This file is part of craft-platforms.
#
# Copyright 2025 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for hypothesis strategies."""

from typing import Dict, Union

import craft_platforms
import pytest
from craft_platforms import _utils
from craft_platforms.test import strategies
from hypothesis import given


@pytest.mark.parametrize("version", strategies._UBUNTU_VERSIONS)
def test_ubuntu_version(version: str):
    if version == "6.06":  # Special case for Dapper Drake
        return
    assert version != "6.04"

    year, month = version.split(".")
    assert month in ("04", "10")
    assert int(year) in range(4, 100)
    assert version != "4.04"


@given(strategies.ubuntu_series())
def test_ubuntu_series(base: craft_platforms.DistroBase) -> None:
    assert base.distribution == "ubuntu"

    assert base.series in strategies._UBUNTU_VERSIONS


@given(strategies.debian_series())
def test_debian_series(base: craft_platforms.DistroBase) -> None:
    assert base.distribution == "debian"

    assert base.series in strategies._DEBIAN_VERSIONS


@given(strategies.ubuntu_lts())
def test_ubuntu_lts(base: craft_platforms.DistroBase) -> None:
    assert base.distribution == "ubuntu"
    assert base.series in strategies._UBUNTU_LTS


@given(strategies.real_distro_base())
def test_real_distro_base(base):
    assert isinstance(base, craft_platforms.DistroBase)
    assert base.distribution in ("debian", "ubuntu")


@given(strategies.any_distro_base())
def test_any_distro_base(base):
    assert isinstance(base, craft_platforms.DistroBase)
    assert isinstance(base.distribution, str)
    assert isinstance(base.series, str)


@given(strategies.build_for_arch_str())
def test_build_for_arch_str(build_for: str):
    if build_for == "all":
        return
    craft_platforms.DebianArchitecture(build_for)


@given(strategies.distro_series_arch_str(strategies.any_distro_base()))
def test_distro_series_arch_str(distro_base_arch_str: str):
    distro_base, arch = distro_base_arch_str.split(":")

    craft_platforms.DistroBase.from_str(distro_base)
    craft_platforms.DebianArchitecture(arch)


@given(strategies.build_on_str(strategies.any_distro_base()))
def test_build_on_str(build_on_str: str):
    base, arch = craft_platforms.parse_base_and_architecture(build_on_str)
    assert base is None or isinstance(base, craft_platforms.DistroBase)
    assert isinstance(arch, craft_platforms.DebianArchitecture)


@given(strategies.build_for_str(strategies.any_distro_base()))
def test_build_for_str(build_for_str: str):
    base, arch = craft_platforms.parse_base_and_architecture(build_for_str)
    assert base is None or isinstance(base, craft_platforms.DistroBase)
    assert arch == "all" or isinstance(arch, craft_platforms.DebianArchitecture)


def check_platform_dict(platform_dict: Union[dict, craft_platforms.PlatformDict]):
    assert platform_dict.keys() == {"build-on", "build-for"}
    build_fors = _utils.vectorize(platform_dict["build-for"])

    for build_on in _utils.vectorize(platform_dict["build-on"]):
        _, build_arch = craft_platforms.parse_base_and_architecture(build_on)
        assert isinstance(build_arch, craft_platforms.DebianArchitecture)
    assert len(build_fors) == 1
    craft_platforms.parse_base_and_architecture(build_fors[0])


@given(strategies.platform_dict(max_build_ons=None))
def test_platform_dicts(platform_dict: dict):
    check_platform_dict(platform_dict)


def check_platforms(platforms: Dict[str, Union[craft_platforms.PlatformDict, None]]):
    for name, value in platforms.items():
        if value is None:
            _, build_arch = craft_platforms.parse_base_and_architecture(name)
            assert isinstance(build_arch, craft_platforms.DebianArchitecture)
        else:
            check_platform_dict(value)


@given(strategies.platform(strategies.any_distro_base()))
def test_platform(platform: Dict[str, Union[craft_platforms.PlatformDict, None]]):
    assert len(platform) == 1
    check_platforms(platform)


@given(strategies.platforms(strategies.any_distro_base()))
def test_platforms(platforms):
    check_platforms(platforms)



# This file is part of starbase.
#
# Copyright 2024 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime

project = "craft-platforms"
author = "Canonical"

copyright = "2023-%s, %s" % (datetime.date.today().year, author)

# region Configuration for canonical-sphinx
ogp_site_url = "https://canonical-craft-platforms.readthedocs-hosted.com/"
ogp_site_name = project
ogp_image = "https://assets.ubuntu.com/v1/253da317-image-document-ubuntudocs.svg"

html_context = {
    "product_page": "github.com/canonical/craft-platforms",
    "github_url": "https://github.com/canonical/craft-platforms",
}

extensions = [
    "canonical_sphinx",
]
# endregion

# region General configuration
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions.extend(
    [
        # "sphinx_toolbox.more_autodoc",
        "sphinx_toolbox.more_autodoc.variables",
        "sphinx.ext.autodoc",  # Must be loaded after more_autodoc
        "sphinx.ext.napoleon",
        "sphinx_autodoc_typehints",
        "sphinx.ext.intersphinx",
        "sphinx.ext.viewcode",
        "sphinx.ext.coverage",
        "sphinx.ext.doctest",
        "sphinx-pydantic",
        "sphinx_toolbox",
    ]
)

exclude_patterns = [
    # Exclude the empty quadrants
    "tutorials/index.rst",
    "how-to/index.rst",
    "explanation/index.rst",
]

# endregion

# region Options for extensions
# Intersphinx extension
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#configuration

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "hypothesis": ("https://hypothesis.readthedocs.io/en/latest", None),
}

# Type annotations config
# add_module_names = True

# Type hints configuration
set_type_checking_flag = True
typehints_fully_qualified = False
always_document_param_types = True
typehints_document_rtype = True

# Autodoc extension configuration
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#configuration
autodoc_member_order = "bysource"
autodoc_default_flags = ["members", "show-inheritance"]
autodoc_typehints_format = "short"

# sphinx-autodoc-typehints configuration
# https://github.com/tox-dev/sphinx-autodoc-typehints?tab=readme-ov-file#options
always_use_bars_union = True
typehints_use_rtype = False
typehints_defaults = "comma"

# More-autodoc configuration
# https://sphinx-toolbox.readthedocs.io/en/stable/extensions/more_autodoc/index.html
overloads_location = "bottom"

# Napoleon configuration
# https://www.sphinx-doc.org/en/master/usage/extensions/napoleon.html
napoleon_attr_annotations = True

# Github config
github_username = "canonical"
github_repository = "craft-platforms"

# endregion



# This file is part of craft-platforms.
#
# Copyright 2024 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Charm-specific module for craft-platforms."""

from ._build import (
    DEFAULT_ARCHITECTURES,
    get_platforms_charm_build_plan,
    get_bases_charm_build_plan,
    get_charm_build_plan,
)


__all__ = [
    "DEFAULT_ARCHITECTURES",
    "get_platforms_charm_build_plan",
    "get_bases_charm_build_plan",
    "get_charm_build_plan",
]



# Copyright 2025 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Integration tests that test end-to-end build planning."""

import pathlib

import craft_platforms
import pytest
import yaml


@pytest.mark.parametrize(
    ("filename"),
    [
        path.name
        for path in (pathlib.Path(__file__).parent / "valid-projects").iterdir()
    ],
)
def test_valid_projects_succeed(filename: str) -> None:
    app_name = filename.partition("-")[0]
    with (pathlib.Path(__file__).parent / "valid-projects" / filename).open() as f:
        project_data = yaml.safe_load(f)

    build_plan = craft_platforms.get_build_plan(app=app_name, project_data=project_data)

    expected = project_data["_build_plan"]

    assert [repr(item) for item in build_plan] == expected



# This file is part of craft-platforms.
#
# Copyright 2024 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for rockcraft builds."""

import itertools

import craft_platforms
import pytest
import pytest_check
from craft_platforms import _errors, rock

SAMPLE_UBUNTU_VERSIONS = ("16.04", "18.04", "20.04", "22.04", "24.04", "24.10", "devel")


@pytest.mark.parametrize(
    ("base", "build_base", "expected_base"),
    [
        *[
            # No special build base
            (f"ubuntu@{version}", None, craft_platforms.DistroBase("ubuntu", version))
            for version in SAMPLE_UBUNTU_VERSIONS
        ],
        *[
            # Always build on a different Ubuntu version
            (
                "ubuntu@00.04",
                f"ubuntu@{version}",
                craft_platforms.DistroBase("ubuntu", version),
            )
            for version in SAMPLE_UBUNTU_VERSIONS
        ],
        # Legacy base strings containing colons
        ("ubuntu:20.04", None, craft_platforms.DistroBase.from_str("ubuntu@20.04")),
        ("ubuntu:22.04", None, craft_platforms.DistroBase.from_str("ubuntu@22.04")),
        ("bare", "ubuntu:20.04", craft_platforms.DistroBase.from_str("ubuntu@20.04")),
        ("bare", "ubuntu:22.04", craft_platforms.DistroBase.from_str("ubuntu@22.04")),
    ],
)
@pytest.mark.parametrize(
    ("platforms", "platform_archs"),
    [
        *[
            pytest.param(
                {architecture.value: None},
                {architecture.value: [(architecture.value, architecture.value)]},
                id=f"implicit-{architecture.value}",
            )
            for architecture in craft_platforms.DebianArchitecture
        ],
        *[
            pytest.param(
                {
                    architecture.value: {
                        "build-on": [architecture.value],
                        "build-for": [architecture.value],
                    },
                },
                {architecture.value: [(architecture.value, architecture.value)]},
                id=f"explicit-{architecture.value}",
            )
            for architecture in craft_platforms.DebianArchitecture
        ],
        *[
            pytest.param(
                {
                    architecture.value: {
                        "build-on": architecture.value,
                        "build-for": architecture.value,
                    },
                },
                {architecture.value: [(architecture.value, architecture.value)]},
                id=f"explicit-scalar-{architecture.value}",
            )
            for architecture in craft_platforms.DebianArchitecture
        ],
        *[
            pytest.param(
                {
                    "my-platform": {
                        "build-on": [
                            arch.value for arch in craft_platforms.DebianArchitecture
                        ],
                        "build-for": [build_for_arch.value],
                    },
                },
                {
                    "my-platform": [
                        (arch.value, build_for_arch.value)
                        for arch in craft_platforms.DebianArchitecture
                    ],
                },
                id=f"build-on-any-for-{build_for_arch.value}",
            )
            for build_for_arch in craft_platforms.DebianArchitecture
        ],
        *[
            pytest.param(
                {
                    "my-platform": {
                        "build-on": [
                            arch.value for arch in craft_platforms.DebianArchitecture
                        ],
                        "build-for": [build_for_arch.value],
                    },
                },
                {
                    "my-platform": [
                        (arch.value, build_for_arch.value)
                        for arch in craft_platforms.DebianArchitecture
                    ],
                },
                id=f"build-on-any-for-scalar-{build_for_arch.value}",
            )
            for build_for_arch in craft_platforms.DebianArchitecture
        ],
    ],
)
def test_build_plans_success(
    check,
    base,
    build_base,
    expected_base,
    platforms,
    platform_archs,
):
    """Shallow test for success on a large number of platform items."""
    build_plan = rock.get_rock_build_plan(
        base=base,
        build_base=build_base,
        platforms=platforms,
    )

    for build_item in build_plan:
        with check():
            assert build_item.build_base == expected_base
        with check():
            assert (build_item.build_on, build_item.build_for) in platform_archs[
                build_item.platform
            ]
    expected_length = len(
        list(
            itertools.chain.from_iterable(
                arch_pairs for arch_pairs in platform_archs.values()
            ),
        ),
    )
    pytest_check.equal(expected_length, len(build_plan))


@pytest.mark.parametrize(
    ("base", "build_base", "platforms", "expected"),
    [
        pytest.param(
            "ubuntu@22.04",
            None,
            {"amd64": None},
            [
                craft_platforms.BuildInfo(
                    "amd64",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DistroBase("ubuntu", "22.04"),
                ),
            ],
            id="jammy-amd64",
        ),
        pytest.param(
            "devel",
            None,
            {"amd64": None},
            [
                craft_platforms.BuildInfo(
                    "amd64",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DistroBase("ubuntu", "devel"),
                ),
            ],
            id="devel-amd64",
        ),
        pytest.param(
            "ubuntu@24.04",
            "ubuntu@22.04",
            {"amd64": None},
            [
                craft_platforms.BuildInfo(
                    "amd64",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DistroBase("ubuntu", "22.04"),
                ),
            ],
            id="jammy-for-noble-amd64",
        ),
        pytest.param(
            "ubuntu@24.04",
            None,
            {
                "my-desktop": {
                    "build-on": ["amd64"],
                    "build-for": ["amd64"],
                },
                "raspi": {"build-on": ["amd64", "arm64"], "build-for": ["arm64"]},
                "some-mainframe-cross-compile": {
                    "build-on": ["amd64", "arm64"],
                    "build-for": ["s390x"],
                },
            },
            [
                craft_platforms.BuildInfo(
                    "my-desktop",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
                craft_platforms.BuildInfo(
                    "raspi",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.ARM64,
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
                craft_platforms.BuildInfo(
                    "raspi",
                    craft_platforms.DebianArchitecture.ARM64,
                    craft_platforms.DebianArchitecture.ARM64,
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
                craft_platforms.BuildInfo(
                    "some-mainframe-cross-compile",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.S390X,
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
                craft_platforms.BuildInfo(
                    "some-mainframe-cross-compile",
                    craft_platforms.DebianArchitecture.ARM64,
                    craft_platforms.DebianArchitecture.S390X,
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
            ],
            id="multiple-builds",
        ),
    ],
)
def test_build_plans_in_depth(base, build_base, platforms, expected):
    """Test the exact build plan for a set of items."""
    actual = rock.get_rock_build_plan(
        base=base,
        build_base=build_base,
        platforms=platforms,
    )

    assert actual == expected


@pytest.mark.parametrize(
    ("base", "error_msg"),
    [
        (
            "invalid-base",
            "Invalid base string 'invalid-base'. Format should be '<distribution>@<series>'",
        ),
    ],
)
def test_build_plans_bad_base(base, error_msg):
    with pytest.raises(ValueError, match=error_msg):
        rock.get_rock_build_plan(base, {"amd64": None})


@pytest.mark.parametrize(
    ("platforms", "error_msg"),
    [
        pytest.param(
            {"my machine": None},
            "Platform name 'my machine' is not a valid Debian architecture. Specify a build-on and build-for.",
            id="invalid-platform-name-no-details",
        ),
        pytest.param(
            {"my machine": {"build-on": ["my machine"], "build-for": ["amd64"]}},
            "'my machine' is not a valid DebianArchitecture",
            id="invalid-architecture-name",
        ),
        pytest.param(
            {"my machine": {"build-on": ["amd64"], "build-for": ["all"]}},
            "platform 'my machine' is invalid",
            id="all",
        ),
    ],
)
def test_build_plans_bad_architecture(platforms, error_msg):
    with pytest.raises(ValueError, match=error_msg):
        rock.get_rock_build_plan("ubuntu@24.04", platforms)


def test_bare_base_no_build_base() -> None:
    """Make sure that an error is raised if base=="bare" but build-base==None"""
    with pytest.raises(_errors.NeedBuildBaseError):
        rock.get_rock_build_plan("bare", platforms={"amd64": None}, build_base=None)



# This file is part of craft-platforms.
#
# Copyright 2024 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for architectures."""

import platform

import pytest
from craft_platforms import DebianArchitecture, DistroBase, parse_base_and_architecture
from craft_platforms.test import strategies
from hypothesis import given


@pytest.mark.parametrize(
    ("given", "expected"),
    [
        pytest.param("AMD64", DebianArchitecture.AMD64, id="amd64-windows"),
        pytest.param("x86_64", DebianArchitecture.AMD64, id="converted"),
        pytest.param("riscv64", DebianArchitecture.RISCV64, id="not-converted"),
    ],
)
def test_debian_architectures_from_machine(given, expected):
    assert DebianArchitecture.from_machine(given) == expected


@pytest.mark.parametrize(
    ("given", "expected"),
    [
        pytest.param(DebianArchitecture.AMD64, "x86_64", id="converted"),
        pytest.param(DebianArchitecture.RISCV64, "riscv64", id="not-converted"),
    ],
)
def test_debian_architecture_to_platform_arch(given, expected):
    assert given.to_platform_arch() == expected


@pytest.mark.parametrize("machine", ["aarch64", "x86_64", "riscv64"])
def test_debian_architecture_from_host(monkeypatch, machine):
    monkeypatch.setattr(platform, "machine", lambda: machine)
    assert DebianArchitecture.from_host().to_platform_arch() == machine


@pytest.mark.parametrize(
    ("given", "expected"),
    [
        (str(DebianArchitecture.AMD64), (None, "amd64")),
        (str(DebianArchitecture.RISCV64), (None, "riscv64")),
        ("all", (None, "all")),
        ("ubuntu@24.04:amd64", (DistroBase("ubuntu", "24.04"), "amd64")),
        ("ubuntu@24.04:riscv64", (DistroBase("ubuntu", "24.04"), "riscv64")),
        ("ubuntu@24.04:all", (DistroBase("ubuntu", "24.04"), "all")),
    ],
)
def test_parse_base_and_architecture(given, expected):
    assert parse_base_and_architecture(given) == expected


def test_parse_base_and_architecture_invalid_arch():
    expected = "'unknown' is not a valid Debian architecture."

    with pytest.raises(ValueError, match=expected):
        parse_base_and_architecture("unknown")


def test_parse_base_and_architecture_invalid_base():
    expected = (
        "Invalid base string 'unknown'. Format should be '<distribution>@<series>'"
    )

    with pytest.raises(ValueError, match=expected):
        parse_base_and_architecture("unknown:riscv64")


@given(
    base=strategies.any_distro_base(),
    arch=strategies.build_for_arch_str(),
)
def test_fuzz_parse_base_and_architecture(base, arch):
    out_base, out_arch = parse_base_and_architecture(f"{base}:{arch}")

    assert out_base == base
    assert out_arch == arch


@pytest.mark.parametrize("deb_arch", list(DebianArchitecture))
def test_debian_architecture_repr(deb_arch: DebianArchitecture):
    """Test that the repr of DebianArchitectures is the repr of their string value."""
    arch_str = deb_arch.value
    assert f"{deb_arch!r}" == f"{arch_str!r}"



# This file is part of craft-platforms.
#
# Copyright 2024 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for error classes."""

from typing import Optional

import pytest
from craft_platforms._errors import CraftPlatformsError

FULL_PLATFORMS_ERROR = CraftPlatformsError(
    message="message",
    details="details",
    resolution="resolution",
    docs_url="https://docs.canonical.com",
    doc_slug="sluggy-mc-slugface",
    logpath_report=False,
    reportable=False,
    retcode=-1,
)


class OtherCraftError(Exception):
    """An exception class that matches the CraftError protocol."""

    def __init__(
        self,
        message: str,
        details: Optional[str] = None,
        resolution: Optional[str] = None,
    ) -> None:
        self.details = details
        self.resolution = resolution
        super().__init__(message)


NON_MATCHING_ERROR = OtherCraftError("message", "details", "resolution")
NON_MATCHING_ERROR.docs_url = "no match"  # pyright: ignore[reportAttributeAccessIssue] # ty: ignore[unresolved-attribute]


class NonError:
    """A non-exception class that matches the CraftError protocol."""

    def __init__(
        self,
        message: str,
        details: Optional[str] = None,
        resolution: Optional[str] = None,
    ) -> None:
        self.args = (message,)
        self.details = details
        self.resolution = resolution


@pytest.mark.parametrize(
    ("this", "that", "expected"),
    [
        (CraftPlatformsError("msg"), CraftPlatformsError("msg"), True),
        (CraftPlatformsError("msg"), OtherCraftError("msg"), True),
        (CraftPlatformsError("msg"), NonError("msg"), False),
        (CraftPlatformsError("msg"), Exception("msg"), False),
        (
            FULL_PLATFORMS_ERROR,
            CraftPlatformsError(
                message="other-message",
                details="details",
                resolution="resolution",
                docs_url="https://docs.canonical.com",
                doc_slug="sluggy-mc-slugface",
                logpath_report=False,
                reportable=False,
                retcode=-1,
            ),
            False,
        ),
        (
            FULL_PLATFORMS_ERROR,
            CraftPlatformsError(
                message="message",
                details="other-details",
                resolution="resolution",
                docs_url="https://docs.canonical.com",
                doc_slug="sluggy-mc-slugface",
                logpath_report=False,
                reportable=False,
                retcode=-1,
            ),
            False,
        ),
        (
            FULL_PLATFORMS_ERROR,
            CraftPlatformsError(
                message="message",
                details="details",
                resolution="other-resolution",
                docs_url="https://docs.canonical.com",
                doc_slug="sluggy-mc-slugface",
                logpath_report=False,
                reportable=False,
                retcode=-1,
            ),
            False,
        ),
        (
            FULL_PLATFORMS_ERROR,
            CraftPlatformsError(
                message="message",
                details="details",
                resolution="resolution",
                docs_url="https://other-docs.canonical.com",
                doc_slug="sluggy-mc-slugface",
                logpath_report=False,
                reportable=False,
                retcode=-1,
            ),
            False,
        ),
        (
            FULL_PLATFORMS_ERROR,
            CraftPlatformsError(
                message="message",
                details="details",
                resolution="resolution",
                docs_url="https://docs.canonical.com",
                doc_slug="other-slug",
                logpath_report=False,
                reportable=False,
                retcode=-1,
            ),
            False,
        ),
        (
            FULL_PLATFORMS_ERROR,
            CraftPlatformsError(
                message="message",
                details="details",
                resolution="resolution",
                docs_url="https://docs.canonical.com",
                doc_slug="sluggy-mc-slugface",
                logpath_report=True,  # Different
                reportable=False,
                retcode=-1,
            ),
            False,
        ),
        (
            FULL_PLATFORMS_ERROR,
            CraftPlatformsError(
                message="message",
                details="details",
                resolution="resolution",
                docs_url="https://docs.canonical.com",
                doc_slug="sluggy-mc-slugface",
                logpath_report=False,
                reportable=True,  # Different
                retcode=-1,
            ),
            False,
        ),
        (
            FULL_PLATFORMS_ERROR,
            CraftPlatformsError(
                message="message",
                details="details",
                resolution="resolution",
                docs_url="https://docs.canonical.com",
                doc_slug="sluggy-mc-slugface",
                logpath_report=False,
                reportable=False,
                retcode=-2,  # Different
            ),
            False,
        ),
        (
            FULL_PLATFORMS_ERROR,
            OtherCraftError("message", "details", "resolution"),
            True,
        ),
        (FULL_PLATFORMS_ERROR, OtherCraftError("nope", "details", "resolution"), False),
        (FULL_PLATFORMS_ERROR, OtherCraftError("message", "nope", "resolution"), False),
        (FULL_PLATFORMS_ERROR, OtherCraftError("message", "details", "nope"), False),
        (FULL_PLATFORMS_ERROR, NON_MATCHING_ERROR, False),
    ],
)
def test_platforms_error_equality(this, that, expected):
    assert (this == that) == expected



# This file is part of craft-platforms.
#
# Copyright 2025 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Hypothesis strategies for testing platforms."""

from typing import Dict, Optional, Union

try:
    from hypothesis import strategies
except ImportError:
    raise ImportError("Test strategies only work if you have hypothesis installed.")

import craft_platforms

_UBUNTU_VERSIONS = (
    "4.10",
    "5.04",
    "5.10",
    "6.06",
    "6.10",
    *(f"{year}.{month}" for year in range(7, 100) for month in ("04", "10")),
)
_UBUNTU_LTS = (
    "6.06",
    "8.04",
    "10.04",
    "12.04",
    "14.04",
    "16.04",
    "20.04",
    "22.04",
    "24.04",
    "26.04",
)
_DEBIAN_VERSIONS = (
    "1.1",
    "1.2",
    "1.3",
    "2.0",
    "2.1",
    "2.2",
    "3.0",
    "3.1",
    "4.0",
    "5.0",
    "6.0",
    "7.0",
    "8.0",
    *(str(version) for version in range(8, 15)),
    "sid",
    "unstable",
)


@strategies.composite
def ubuntu_series(draw: strategies.DrawFn) -> craft_platforms.DistroBase:
    """Generate Ubuntu series as DistroBase objects.

    This strategy includes both LTS and interim releases, as well as the special
    series "devel" that defines the currently under development Ubuntu release.
    """
    return craft_platforms.DistroBase(
        distribution="ubuntu", series=draw(strategies.sampled_from(_UBUNTU_VERSIONS))
    )


@strategies.composite
def ubuntu_lts(draw: strategies.DrawFn) -> craft_platforms.DistroBase:
    """Generate LTS Ubuntu series as DistroBase objects.

    This includes Ubuntu LTS releases all the way back to Ubuntu 6.06 (Dapper Drake)
    and may include future Ubuntu LTS releases as well.
    """
    return craft_platforms.DistroBase(
        distribution="ubuntu", series=draw(strategies.sampled_from(_UBUNTU_LTS))
    )


@strategies.composite
def debian_series(draw: strategies.DrawFn) -> craft_platforms.DistroBase:
    """Generate Debian releases as DistroBase objects.

    This includes all stable Debian releases by version number (up to Debian 14) as
    well as the special values ``"sid"`` and ``"unstable"``. Because ``"testing"``,
    ``"stable"`` and ``"oldstable"`` refer to different releases given the current
    time, they are not included.
    """
    return craft_platforms.DistroBase(
        distribution="debian", series=draw(strategies.sampled_from(_DEBIAN_VERSIONS))
    )


def real_distro_base() -> strategies.SearchStrategy[craft_platforms.DistroBase]:
    """Generate DistroBase objects of real (possibly future) Linux distributions.

    This is the default strategy when generating ``DistroBase`` objects.
    """
    return strategies.one_of(ubuntu_series(), debian_series())


@strategies.composite
def any_distro_base(draw: strategies.DrawFn) -> craft_platforms.DistroBase:
    """Generate any ``DistroBase`` that vaguely looks as expected."""
    return craft_platforms.DistroBase(
        distribution=draw(
            strategies.text("abcdefghijklmnopqrstuvwxyz-", min_size=1).filter(
                lambda x: not x.startswith("-") and not x.endswith("-")
            )
        ),
        series=str(
            draw(
                strategies.one_of(
                    strategies.sampled_from(
                        ["devel", "dev", "unstable", "testing", "sid", "tumbleweed"]
                    ),
                    strategies.floats(),
                )
            )
        ),
    )


def build_on_arch_str() -> strategies.SearchStrategy[str]:
    """Generate valid ``build-on`` architecture strings.

    This is essentially the string forms of :class:`~craft_platforms.DebianArchitecture`
    """
    return strategies.sampled_from(
        [arch.value for arch in craft_platforms.DebianArchitecture]
    )


def build_for_arch_str() -> strategies.SearchStrategy[str]:
    """Generate valid ``build-for`` architecture strings.

    These are the string forms of any :class:`~craft_platforms.DebianArchitecture`
    plus the special value ``"all"``. This should be used if you want only architecture
    strings without the multi-base format.
    """
    return strategies.sampled_from(
        ["all", *(arch.value for arch in craft_platforms.DebianArchitecture)]
    )


@strategies.composite
def distro_series_arch_str(
    draw: strategies.DrawFn,
    distro_base: strategies.SearchStrategy[craft_platforms.DistroBase],
    arch: Optional[strategies.SearchStrategy[str]] = None,
) -> str:
    """Generate ``distro@series:arch`` strings.

    :param distro_base: A strategy for generating :class:`~craft_platforms.DistroBase`
        objects.
    :param arch: A strategy for generating Debian architecture name strings. Defaults
        to the values of :class:`~craft_platforms.DebianArchitecture`. Use
        :func:`build_for_arch_str` here to include ``all`` as an architecture.
    """
    if arch is None:
        arch = strategies.sampled_from(craft_platforms.DebianArchitecture).map(
            lambda arch: arch.value
        )
    distro_str = str(draw(distro_base))
    arch_str = draw(arch)
    return f"{distro_str}:{arch_str}"


def build_on_str(
    distro_base: strategies.SearchStrategy[craft_platforms.DistroBase],
) -> strategies.SearchStrategy[str]:
    """Generate valid ``build-on`` strings for platforms.

    This strategy includes both architecture values and the values used for multi-base
    platform definitions.

    :param distro_base: A strategy for generating :class:`~craft_platforms.DistroBase`
        objects.
    """
    return strategies.one_of(
        strategies.sampled_from(craft_platforms.DebianArchitecture).map(
            lambda arch: arch.value
        ),
        distro_series_arch_str(distro_base),
    )


def build_for_str(
    distro_base: strategies.SearchStrategy[craft_platforms.DistroBase],
) -> strategies.SearchStrategy[str]:
    """Generate valid ``build-for`` strings for platforms.

    This strategy includes both architecture values (and ``"all"``) as well as valid
    values for multi-base platform definitions. Use :func:`build_for_arch_str` if you
    only want architecture strings or ``"all"``.

    :param distro_base: A strategy for generating :class:`~craft_platforms.DistroBase`
        objects.
    """
    return strategies.one_of(
        strategies.sampled_from(
            ["all", *(arch.value for arch in craft_platforms.DebianArchitecture)]
        ),
        distro_series_arch_str(
            distro_base=distro_base,
            arch=strategies.sampled_from(
                ["all", *(arch.value for arch in craft_platforms.DebianArchitecture)]
            ),
        ),
    )


@strategies.composite
def platform_dict(
    draw: strategies.DrawFn,
    build_ons: Optional[strategies.SearchStrategy[str]] = None,
    build_fors: Optional[strategies.SearchStrategy[str]] = None,
    max_build_ons: Optional[int] = 8,
) -> craft_platforms.PlatformDict:
    """Generate platform dictionaries.

    .. caution::

        The platform dictionaries generated here only limited by the rules of what
        Craft Platforms considers a valid platform dictionary. Most applications have
        far stricter rules for their platform dictionaries, meaning this strategy
        is almost guaranteed to generate invalid dictionaries for an application.
        If you need ``PlatformDict`` objects that are valid for a particular
        application, it is better to build an app-specific strategy.

    :param build_ons: A strategy for creating ``build-on`` strings. Defaults to
        :func:`build_on_str`
    :param build_fors: A strategy for creating ``build-for`` strings. Defaults to
        :func:`build_for_str`
    :param max_build_ons: The maximum length of the ``build-on`` list.
    """
    if build_ons is None:
        build_ons = build_on_str(real_distro_base())
    if build_fors is None:
        build_fors = build_for_str(real_distro_base())
    return {
        "build-on": draw(
            strategies.one_of(
                strategies.lists(
                    build_ons,
                    min_size=1,
                    max_size=max_build_ons,
                    unique_by=lambda x: tuple(sorted(x)),
                ),
                build_ons,
            )
        ),
        "build-for": draw(
            strategies.one_of(
                build_fors,
                strategies.lists(build_fors, min_size=1, max_size=1),  # or list of them
            )
        ),
    }


def platform(
    distro_base: strategies.SearchStrategy[craft_platforms.DistroBase],
    shorthand_keys: Optional[strategies.SearchStrategy[str]] = None,
    values: Union[strategies.SearchStrategy[craft_platforms.PlatformDict], None] = None,
) -> strategies.SearchStrategy[
    Union[Dict[str, None], Dict[str, craft_platforms.PlatformDict]]
]:
    """Generate a single platform in a dictionary.

    The generated platform dictionary could either be of the shorthand form
    (``{"[<distro>@<series>:]<arch>": None}``) or of the longhand form.

    .. caution::

        The default values used here generate platform dictionaries that are not
        valid for most applications.

    :param distro_base: A strategy that generates :class:`~craft_platforms.DistroBase``
        objects.
    :param shorthand_keys: A strategy that generates valid shorthand keys. These could
        be architecture names or ``[<distro>@<base>:]<arch>`` strings (the default).
    :param values: A search strategy that generates ``PlatformDict`` objects. If not
        set, uses :func:`platform_dicts` with default arguments.
    """
    if shorthand_keys is None:
        shorthand_keys = build_on_str(distro_base)
    if values is None:
        values = platform_dict()
    return strategies.one_of(
        strategies.dictionaries(
            keys=shorthand_keys,
            values=strategies.none(),
            min_size=1,
            max_size=1,
        ),
        strategies.dictionaries(
            keys=strategies.text(min_size=1).filter(lambda x: x not in ("@", ":")),
            values=values,
            min_size=1,
            max_size=1,
        ),
    )


@strategies.composite
def platforms(
    draw: strategies.DrawFn,
    distro_base: strategies.SearchStrategy[craft_platforms.DistroBase],
    shorthand_keys: Optional[strategies.SearchStrategy[str]] = None,
    values: Optional[strategies.SearchStrategy[craft_platforms.PlatformDict]] = None,
    min_size: int = 1,
    max_size: Union[int, None] = 8,
) -> Dict[str, Optional[craft_platforms.PlatformDict]]:
    """Generate platforms dictionaries.

    Generates platforms dictionaries with a size in the given range.

    .. warning:: This function can be slow, especially when generating large dictionaries.

    :param distro_base: A strategy that generates :class:`~craft_platforms.DistroBase``
        objects.
    :param shorthand_keys: A strategy that generates valid shorthand keys. These could
        be architecture names or ``[<distro>@<base>:]<arch>`` strings (the default).
    :param values: A search strategy that generates ``PlatformDict`` objects. If not
        set, uses :func:`platform_dicts` with default arguments.
    :param min_size: The minimum size of the dictionary. Must be ``>= 1``.
    :param max_size: The maximum size of the dictionary, or None for unbounded.
    """
    list_of_platforms = draw(
        strategies.lists(
            platform(distro_base, shorthand_keys, values),
            min_size=min_size,
            max_size=max_size,
            unique_by=lambda p_dict: tuple(p_dict.keys()),
        )
    )
    platforms: Dict[str, Optional[craft_platforms.PlatformDict]] = {}
    for p in list_of_platforms:
        platforms.update(p)
    return platforms



# This file is part of craft-platforms.
#
# Copyright 2024 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Platform related models."""

import itertools
import typing
from typing import Dict, List, Optional, Sequence, Tuple, Union

import annotated_types
from typing_extensions import Annotated

from craft_platforms import _architectures, _buildinfo, _distro, _errors, _utils

PlatformDict = typing.TypedDict(
    "PlatformDict",
    {
        "build-on": Union[Sequence[str], str],
        "build-for": Union[Annotated[Sequence[str], annotated_types.Len(1)], str],
    },
)
"""The platforms where an artifact is built and where the resulting artifact runs."""


Platforms = Dict[Union[_architectures.DebianArchitecture, str], Optional[PlatformDict]]
"""A mapping of platforms names to ``PlatformDicts``.

A ``PlatformDict`` is not required if the platform name is a supported Debian architecture.
"""


def get_platforms_build_plan(
    base: Union[str, _distro.DistroBase],
    platforms: Platforms,
    build_base: Optional[str] = None,
) -> Sequence[_buildinfo.BuildInfo]:
    """Generate the build plan for a platforms-based artefact."""
    if isinstance(base, _distro.DistroBase):
        distro_base = base
    else:
        distro_base = _distro.DistroBase.from_str(build_base or base)
    build_plan: List[_buildinfo.BuildInfo] = []

    for platform_name, platform in platforms.items():
        if platform is None:
            # This is a workaround for Python 3.10.
            # In python 3.12+ we can just check:
            # `if platform_name not in _architectures.DebianArchitecture`
            try:
                architecture = _architectures.DebianArchitecture(platform_name)
            except ValueError:
                raise _errors.InvalidPlatformNameError(
                    f"Platform name {platform_name!r} is not a valid Debian architecture. "
                    "Specify a build-on and build-for.",
                ) from None
            build_plan.append(
                _buildinfo.BuildInfo(
                    platform=platform_name,
                    build_on=architecture,
                    build_for=architecture,
                    build_base=distro_base,
                ),
            )
        else:
            for build_on, build_for in itertools.product(
                _utils.vectorize(platform["build-on"]),
                _utils.vectorize(platform.get("build-for", [platform_name])),
            ):
                build_plan.append(
                    _buildinfo.BuildInfo(
                        platform=platform_name,
                        build_on=_architectures.DebianArchitecture(build_on),
                        build_for=(
                            "all"
                            if build_for == "all"
                            else _architectures.DebianArchitecture(build_for)
                        ),
                        build_base=distro_base,
                    ),
                )

    build_for_archs = {info.build_for for info in build_plan}
    if "all" in build_for_archs:
        platforms_with_all = {
            info.platform for info in build_plan if info.build_for == "all"
        }
        if len(platforms_with_all) > 1:
            raise _errors.AllSinglePlatformError(platforms_with_all)
        if len(build_for_archs) > 1:
            raise _errors.AllOnlyBuildError(platforms_with_all)

    return build_plan


def parse_base_and_name(platform_name: str) -> Tuple[Optional[_distro.DistroBase], str]:
    """Get the platform name and optional base from a platform name.

    The platform name may have an optional base prefix as '[<base>:]<platform>'.

    :param platform_name: The name of the platform.

    :returns: A tuple of the DistroBase and the platform name.

    :raises ValueError: If the base is invalid.
    """
    if ":" in platform_name:
        base_str, _, name = platform_name.partition(":")
        base = _distro.DistroBase.from_str(base_str)
    else:
        base = None
        name = platform_name

    return base, name



# This file is part of craft-platforms.
#
# Copyright 2024 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Charmcraft-specific platforms information."""

import itertools
from typing import Any, Collection, Dict, Iterable, List, Optional, Sequence

from craft_platforms import (
    _architectures,
    _buildinfo,
    _distro,
    _errors,
    _platforms,
    _utils,
)

DEFAULT_ARCHITECTURES: Collection[_architectures.DebianArchitecture] = (
    _architectures.DebianArchitecture.AMD64,
    _architectures.DebianArchitecture.ARM64,
    _architectures.DebianArchitecture.PPC64EL,
    _architectures.DebianArchitecture.RISCV64,
    _architectures.DebianArchitecture.S390X,
)
"""Default architectures for building a charm.

If no platforms are defined, the charm will be built on and for these architectures.
"""


def _validate_base_definition(
    base: Optional[str],
    build_base: Optional[str],
    platform_name: Optional[str],
    platform: Optional[_platforms.PlatformDict],
) -> None:
    """Validate that a base is defined correctly in the data used to create a build.

    The rules are:
     - a base must be defined in only one place
     - each platform must build on and build for the same base

    :raises ValueError: If the base is not defined correctly in the build data.
    """
    if not (platform_name or base or build_base):
        raise _errors.RequiresBaseError(
            message="No base, build-base, or platforms are declared.",
            resolution="Declare a base or build-base.",
        )

    if not platform_name:
        return

    # validate base defined in the platform name
    platform_base, _ = _platforms.parse_base_and_name(platform_name=platform_name)

    if platform:
        if platform_base:
            raise _errors.InvalidMultiBaseError(
                message=(
                    f"Platform {platform_name!r} declares a base in the platform's "
                    "name and declares 'build-on' and 'build-for' entries."
                ),
                resolution=(
                    "Either remove the base from the platform's name or remove the "
                    "'build-on' and 'build-for' entries for the platform."
                ),
            )
        # create a set of the bases defined in the build-on and build-for entries
        bases = set()
        for entry in [
            *_utils.vectorize(platform["build-on"]),
            *_utils.vectorize(platform["build-for"]),
        ]:
            distro_base, _ = _architectures.parse_base_and_architecture(arch=entry)
            bases.add(str(distro_base) if distro_base else None)

        if len(bases) == 0:
            # an empty set means no bases are defined
            build_on_for_base = None
        elif len(bases) == 1:
            # a set with one element means the same base was defined for all entries
            build_on_for_base = next(iter(bases))
        else:
            # otherwise there are multiple bases defined or some entries missing bases
            raise _errors.InvalidMultiBaseError(
                message=(
                    f"Platform {platform_name!r} has mismatched bases in the 'build-on' "
                    "and 'build-for' entries."
                ),
                resolution=(
                    "Use the same base for all 'build-on' and 'build-for' entries for "
                    "the platform."
                ),
            )
    else:
        build_on_for_base = None

    if (platform_base or build_on_for_base) and (base or build_base):
        raise _errors.InvalidMultiBaseError(
            message=f"Platform {platform_name!r} declares a base and a top-level base "
            "or build-base is declared.",
            resolution=(
                "Remove the base from the platform's name or remove the top-level base "
                "or build-base."
            ),
        )

    if not (platform_base or build_on_for_base) and not (base or build_base):
        raise _errors.RequiresBaseError(
            message=(
                "No base or build-base is declared and no base is declared "
                "in the platforms section."
            ),
            resolution="Declare a base or build-base.",
        )


def _get_base_from_build_data(
    base: Optional[str],
    build_base: Optional[str],
    platform_name: Optional[str],
    platform: Optional[_platforms.PlatformDict],
) -> _distro.DistroBase:
    """Get the base from a data used to create a build.

    :returns: The base to use for a build.

    :raises ValueError: If the base is not defined correctly in the build data.
    """
    _validate_base_definition(
        base=base,
        build_base=build_base,
        platform_name=platform_name,
        platform=platform,
    )

    if build_base:
        return _distro.DistroBase.from_str(build_base)

    if base:
        return _distro.DistroBase.from_str(base)

    if platform_name:
        platform_base, _ = _platforms.parse_base_and_name(platform_name=platform_name)
        if platform_base:
            return platform_base

        # build-on and build-for entries all have the same base, so we only
        # need to check one of them
        if platform:
            build_for_base, _ = _architectures.parse_base_and_architecture(
                arch=_utils.vectorize(platform["build-for"])[0]
            )
            if build_for_base:
                return build_for_base

    # if this is raised, then the validator is not working correctly
    raise ValueError("Could not determine the base for the build.")


def get_platforms_charm_build_plan(
    base: Optional[str],
    platforms: Optional[_platforms.Platforms],
    build_base: Optional[str] = None,
) -> Sequence[_buildinfo.BuildInfo]:
    """Generate the build plan for a platforms-based charm.

    Platforms-based charms are charms that don't use the deprecated ``bases``
    field in their ``charmcraft.yaml``.

    Multi-base recipes are supported. A multi-base recipe defines the base
    within the ``platform`` field instead of defining ``base`` and
    ``build-base``. For each platform, the base is either prefixed to the
    platform name or prefixed to every ``build-on`` and ``build-for` entry.
    In both cases, the prefixed base is delimited with a colon (``<base>:``).

    :param base: The run-time environment for the charm, formatted  as
      ``distribution@series``. If the ``build-base`` is unset, then the ``base``
      determines the build environment.
    :param build_base: The build environment to using when building the charm,
      formatted as ``distribution@series``.
    :param platforms: The mapping of platform names to ``PlatformDicts``. If
      the ``base`` and ``build-base`` are unset, then the base must be defined
      in the platforms.

    :raises ValueError: If the build plan can't be created due to invalid base
      and platform definitions.

    :returns: A build plan describing the environments where the charm can build
      and where the charm can run.
    """
    if platforms is None:
        distro_base = _get_base_from_build_data(
            base=base,
            build_base=build_base,
            platform_name=None,
            platform=None,
        )

        # If no platforms are specified, build for all default architectures without
        # an option of cross-compiling.
        return [
            _buildinfo.BuildInfo(
                platform=arch.value,
                build_on=arch,
                build_for=arch,
                build_base=distro_base,
            )
            for arch in DEFAULT_ARCHITECTURES
        ]
    build_plan: List[_buildinfo.BuildInfo] = []
    for platform_name, platform in platforms.items():
        distro_base = _get_base_from_build_data(
            base=base,
            build_base=build_base,
            platform_name=platform_name,
            platform=platform,
        )

        if platform is None:
            _, arch_str = _platforms.parse_base_and_name(platform_name)

            # This is a workaround for Python 3.10.
            # In python 3.12+ we can just check:
            # `if platform_name not in _architectures.DebianArchitecture`
            try:
                arch = _architectures.DebianArchitecture(arch_str)
            except ValueError:
                raise ValueError(
                    f"Platform name {platform_name!r} is not a valid Debian architecture. "
                    "Specify a build-on and build-for.",
                ) from None

            build_plan.append(
                _buildinfo.BuildInfo(
                    platform=platform_name,
                    build_on=arch,
                    build_for=arch,
                    build_base=distro_base,
                ),
            )
        else:
            for build_on, build_for in itertools.product(
                _utils.vectorize(platform["build-on"]),
                _utils.vectorize(platform["build-for"]),
            ):
                _, build_on_arch = _architectures.parse_base_and_architecture(
                    arch=build_on
                )
                if build_on_arch == "all":
                    raise ValueError(
                        f"Platform {platform_name!r} has an invalid 'build-on' entry of 'all'."
                    )

                _, build_for_arch = _architectures.parse_base_and_architecture(
                    arch=build_for
                )

                build_plan.append(
                    _buildinfo.BuildInfo(
                        platform=platform_name,
                        build_on=build_on_arch,
                        build_for=build_for_arch,
                        build_base=distro_base,
                    ),
                )

    return build_plan


def _gen_build_plan_for_base(base: Dict[str, Any]) -> Iterable[_buildinfo.BuildInfo]:
    if "build-on" not in base:
        base = {"build-on": [base], "run-on": [base]}

    for build_base in base["build-on"]:
        build_archs = build_base.get("architectures", DEFAULT_ARCHITECTURES)
        for run_base, build_arch in itertools.product(base["run-on"], build_archs):
            run_archs = run_base.get("architectures", [build_arch])
            run_archs_str = "-".join(run_archs)
            yield _buildinfo.BuildInfo(
                f"{build_base['name']}-{build_base['channel']}-{run_archs_str}",
                build_on=_architectures.DebianArchitecture(build_arch),
                build_for=run_archs[0],
                build_base=_distro.DistroBase(
                    build_base["name"], build_base["channel"]
                ),
            )


def get_bases_charm_build_plan(
    bases: Sequence[Dict[str, Any]],
) -> Sequence[_buildinfo.BuildInfo]:
    """Get a build plan for a legacy "bases" based charm."""
    plan: List[_buildinfo.BuildInfo] = []
    for base in bases:
        plan.extend(_gen_build_plan_for_base(base))
    return plan


def get_charm_build_plan(
    project_data: Dict[str, Any],
) -> Sequence[_buildinfo.BuildInfo]:
    if "platforms" in project_data:
        return get_platforms_charm_build_plan(
            base=project_data.get("base"),
            build_base=project_data.get("build-base"),
            platforms=project_data.get(
                "platforms",
            ),
        )
    if "bases" in project_data:
        return get_bases_charm_build_plan(project_data["bases"])
    raise NotImplementedError("Unknown charm type with no bases or platforms.")



# This file is part of craft-platforms.
#
# Copyright 2024 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Rockcraft-specific platforms information."""

from typing import Optional, Sequence

from craft_platforms import _buildinfo, _errors, _platforms

_LEGACY_BASES_MAP = {
    "ubuntu:20.04": "ubuntu@20.04",
    "ubuntu:22.04": "ubuntu@22.04",
}


def get_rock_build_plan(
    base: str,
    platforms: _platforms.Platforms,
    build_base: Optional[str] = None,
) -> Sequence[_buildinfo.BuildInfo]:
    """Generate the build plan for a rock.

    This function uses the default build planner, but filters it to prevent the use of
    ``build-for: all``

    :param base: the rock base (e.g. ``'ubuntu@24.04'``)
    :param platforms: the platforms structure in ``rockcraft.yaml``
    :param build_base: the build base, if provided in ``rockcraft.yaml``.
    :raises NeedsBuildBaseError: If base is bare and no build base is specified
    """
    # Bare bases require a build_base
    if base == "bare" and build_base is None:
        raise _errors.NeedBuildBaseError(base=base)

    if base in _LEGACY_BASES_MAP:
        base = _LEGACY_BASES_MAP[base]
    if build_base in _LEGACY_BASES_MAP:
        build_base = _LEGACY_BASES_MAP[build_base]

    for name, platform in platforms.items():
        if platform and "all" in platform.get("build-for", []):
            raise _errors.InvalidPlatformError(
                name,
                details="Rockcraft cannot build platform-independent images.",
                resolution="Replace 'build-for: [all]' with a valid architecture",
            )
    return _platforms.get_platforms_build_plan(base, platforms, build_base)



# This file is part of craft-platforms.
#
# Copyright 2024 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for snap-specific platforms functions."""

from typing import Dict, List, Optional, Union

import craft_platforms
import pytest
import pytest_check
from craft_platforms import _errors
from craft_platforms._distro import DistroBase
from craft_platforms.snap import _build
from hypothesis import given, strategies

PLATFORMS_AND_EXPECTED_ARCHES = [
    *[
        pytest.param(
            {architecture.value: None},
            {architecture.value: [(architecture.value, architecture.value)]},
            id=f"implicit-{architecture.value}",
        )
        for architecture in craft_platforms.DebianArchitecture
    ],
    *[
        pytest.param(
            {
                architecture.value: {
                    "build-on": [architecture.value],
                    "build-for": [architecture.value],
                },
            },
            {architecture.value: [(architecture.value, architecture.value)]},
            id=f"explicit-{architecture.value}",
        )
        for architecture in craft_platforms.DebianArchitecture
    ],
    *[
        pytest.param(
            {
                architecture.value: {
                    "build-on": architecture.value,
                    "build-for": architecture.value,
                },
            },
            {architecture.value: [(architecture.value, architecture.value)]},
            id=f"explicit-scalar-{architecture.value}",
        )
        for architecture in craft_platforms.DebianArchitecture
    ],
    *[
        pytest.param(
            {
                "my-platform": {
                    "build-on": [
                        arch.value for arch in craft_platforms.DebianArchitecture
                    ],
                    "build-for": [build_for_arch.value],
                },
            },
            {
                "my-platform": [
                    (arch.value, build_for_arch.value)
                    for arch in craft_platforms.DebianArchitecture
                ],
            },
            id=f"build-on-any-for-{build_for_arch.value}",
        )
        for build_for_arch in craft_platforms.DebianArchitecture
    ],
    *[
        pytest.param(
            {
                "my-platform": {
                    "build-on": [
                        arch.value for arch in craft_platforms.DebianArchitecture
                    ],
                    "build-for": build_for_arch.value,
                },
            },
            {
                "my-platform": [
                    (arch.value, build_for_arch.value)
                    for arch in craft_platforms.DebianArchitecture
                ],
            },
            id=f"build-on-any-for-scalar-{build_for_arch.value}",
        )
        for build_for_arch in craft_platforms.DebianArchitecture
    ],
]

SNAP_TYPES_WITH_BASE = ["gadget", "app", None, "some-future-type"]
SNAP_TYPES = [*SNAP_TYPES_WITH_BASE, *_build.SNAP_TYPES_WITHOUT_BASE]


@given(strategies.integers(min_value=16, max_value=98).filter(lambda x: x % 2 == 0))
def test_core_base_regex_match(version):
    assert _build.CORE_BASE_REGEX.match(f"core{version}")


@given(
    strategies.one_of(
        strategies.integers(min_value=99),
        strategies.integers(min_value=16, max_value=98).filter(lambda x: x % 2 != 0),
        strategies.integers(max_value=15),
    ),
)
def test_core_base_regex_version_non_match(version):
    assert not _build.CORE_BASE_REGEX.match(f"core{version}")


@given(strategies.text().filter(lambda s: not s.startswith("core")))
def test_core_base_regex_non_match(string):
    assert not _build.CORE_BASE_REGEX.match(string)


@pytest.mark.parametrize(
    ("base", "expected"),
    [
        ("core", craft_platforms.DistroBase("ubuntu", "16.04")),
        ("devel", craft_platforms.DistroBase("ubuntu", "devel")),
        *(
            (f"core{n}", craft_platforms.DistroBase("ubuntu", f"{n}.04"))
            for n in range(16, 99, 2)
        ),
        *(
            (f"ubuntu@{n}.04", craft_platforms.DistroBase("ubuntu", f"{n}.04"))
            for n in range(16, 99, 2)
        ),
    ],
)
def test_get_distro_base_from_core_base_success(base, expected):
    assert _build.get_distro_base_from_core_base(base) == expected


@pytest.mark.parametrize("base", ["bare", "flared"])
def test_get_distro_base_from_core_base_error(base: str):
    with pytest.raises(ValueError, match="Invalid base string .+. Format should be"):
        _build.get_distro_base_from_core_base(base)


@pytest.mark.parametrize(
    ("base", "build_base", "expected"),
    [
        ("core", None, DistroBase("ubuntu", "16.04")),
        *(
            (f"core{n}", None, DistroBase("ubuntu", f"{n}.04"))
            for n in range(16, 99, 2)
        ),
        ("core", "devel", DistroBase("ubuntu", "devel")),
        *(
            (f"core{n}", "devel", DistroBase("ubuntu", "devel"))
            for n in range(16, 99, 2)
        ),
        *(
            ("bare", f"core{n}", DistroBase("ubuntu", f"{n}.04"))
            for n in range(16, 99, 2)
        ),
        *(
            ("core22-desktop", f"core{n}", DistroBase("ubuntu", f"{n}.04"))
            for n in range(16, 99, 2)
        ),
        *(
            ("some-arbitrary-base-name", f"core{n}", DistroBase("ubuntu", f"{n}.04"))
            for n in range(16, 99, 2)
        ),
        *(
            ("some-arbitrary-base-name", f"debian@{n}", DistroBase("debian", str(n)))
            for n in range(2, 12)
        ),
        ("bare", "ubuntu@23.10", DistroBase("ubuntu", "23.10")),
        ("core24-alex", "core22", DistroBase("ubuntu", "22.04")),
        ("core24-alex", "devel", DistroBase("ubuntu", "devel")),
        ("arbitrary-base-name", "opensuse@15", DistroBase("opensuse", "15")),
    ],
)
@pytest.mark.parametrize("snap_type", SNAP_TYPES)
def test_get_snap_base_general_success(base, build_base, snap_type, expected):
    assert (
        _build.get_snap_base(base=base, build_base=build_base, snap_type=snap_type)
        == expected
    )


@pytest.mark.parametrize(
    ("base", "build_base", "snap_type", "expected"),
    [
        ("core24", "ubuntu@24.04", "kernel", DistroBase("ubuntu", "24.04")),
        ("arbitrary", "debian@11", "kernel", DistroBase("debian", "11")),
    ],
)
def test_get_snap_base_examples_success(base, build_base, snap_type, expected):
    assert (
        _build.get_snap_base(base=base, build_base=build_base, snap_type=snap_type)
        == expected
    )


@pytest.mark.parametrize("snap_type", SNAP_TYPES_WITH_BASE)
@pytest.mark.parametrize(
    ("base", "match"),
    [
        pytest.param(
            "bare",
            "must declare a 'build-base' if 'base' does not match 'coreXY'",
            id="bare",
        ),
        pytest.param(
            "some-arbitrary-base",
            "must declare a 'build-base' if 'base' does not match 'coreXY'",
            id="some-arbitrary-base",
        ),
        pytest.param(
            None,
            "snaps of type [a-zN'-]+ require a 'base'",
            id="no-base",
        ),
    ],
)
def test_get_snap_base_no_build_base_requires_base_error(base, snap_type, match):
    with pytest.raises(_errors.CraftPlatformsError, match=match):
        _build.get_snap_base(base=base, build_base=None, snap_type=snap_type)


@pytest.mark.parametrize("snap_type", _build.SNAP_TYPES_WITHOUT_BASE)
@pytest.mark.parametrize(
    ("base", "match"),
    [
        pytest.param(
            "bare",
            "must declare a 'build-base' if 'base' does not match 'coreXY'",
            id="bare",
        ),
        pytest.param(
            "some-arbitrary-base",
            "must declare a 'build-base' if 'base' does not match 'coreXY'",
            id="some-arbitrary-base",
        ),
        pytest.param(
            None,
            "'[a-z]+' snaps require a 'build-base' if no 'base' is declared",
            id="no-base",
        ),
    ],
)
def test_get_snap_base_no_build_base_ether_or(base, snap_type, match):
    with pytest.raises(_errors.CraftPlatformsError, match=match):
        _build.get_snap_base(base=base, build_base=None, snap_type=snap_type)


@pytest.mark.parametrize(
    ("base", "build_base", "snap_type", "match"),
    [
        (
            None,
            "devel",
            "kernel",
            "non-base snaps require a 'base' if 'build-base' is 'devel",
        ),
        (None, "flared", "base", "build-base 'flared' is unknown or invalid"),
        (
            "core24",
            "ubuntu@24.04",
            "blah",
            "non-kernel snaps cannot use 'base: coreXY' and arbitrary build-bases",
        ),
    ],
)
def test_get_snap_base_specific_errors(base, build_base, snap_type, match):
    with pytest.raises(_errors.CraftPlatformsError, match=match):
        _build.get_snap_base(base=base, build_base=build_base, snap_type=snap_type)


@pytest.mark.parametrize(
    ("base", "build_base", "expected_base"),
    [
        ("core", None, craft_platforms.DistroBase("ubuntu", "16.04")),
        ("bare", "core", craft_platforms.DistroBase("ubuntu", "16.04")),
        *(
            (f"core{n}", None, craft_platforms.DistroBase("ubuntu", f"{n}.04"))
            for n in (16, 18, 20, 22, 24, 26)
        ),
        *(
            ("bare", f"core{n}", craft_platforms.DistroBase("ubuntu", f"{n}.04"))
            for n in (16, 18, 20, 22, 24, 26)
        ),
        *(
            (
                f"core{n}-desktop",
                f"core{n}",
                craft_platforms.DistroBase("ubuntu", f"{n}.04"),
            )
            for n in (16, 18, 20, 22, 24, 26)
        ),
        *(
            pytest.param(
                f"core{n}",
                "devel",
                craft_platforms.DistroBase("ubuntu", "devel"),
                id=f"on-devel-for-core{n}",
            )
            for n in (16, 18, 20, 22, 24, 26)
        ),
    ],
)
@pytest.mark.parametrize(
    ("platforms", "expected_archs"),
    PLATFORMS_AND_EXPECTED_ARCHES,
)
@pytest.mark.parametrize("snap_type", ["app", "core", "gadget", "kernel", None])
def test_get_platforms_snap_build_plan_success(
    check,
    base: str,
    build_base: Optional[str],
    expected_base: craft_platforms.DistroBase,
    platforms: Union[craft_platforms.Platforms, None],
    expected_archs: Dict[str, List[str]],
    snap_type: Optional[str],
):
    build_plan = _build.get_platforms_snap_build_plan(
        base,
        platforms=platforms,
        build_base=build_base,
        snap_type=snap_type,
    )

    for build_item in build_plan:
        with check():
            assert build_item.build_base == expected_base
        with check():
            assert (build_item.build_on, build_item.build_for) in expected_archs[
                build_item.platform
            ]


@pytest.mark.parametrize(
    ("build_base", "expected_base"),
    [
        ("core", craft_platforms.DistroBase("ubuntu", "16.04")),
        *(
            (f"core{n}", craft_platforms.DistroBase("ubuntu", f"{n}.04"))
            for n in (16, 18, 20, 22, 24, 26)
        ),
        ("devel", craft_platforms.DistroBase("ubuntu", "devel")),
    ],
)
@pytest.mark.parametrize(
    ("platforms", "expected_archs"),
    PLATFORMS_AND_EXPECTED_ARCHES,
)
def test_get_platforms_snap_build_plan_base_snap_success(
    check,
    build_base: str,
    expected_base: craft_platforms.DistroBase,
    platforms: Union[craft_platforms.Platforms, None],
    expected_archs: Dict[str, List[str]],
):
    build_plan = _build.get_platforms_snap_build_plan(
        base=None,
        platforms=platforms,
        build_base=build_base,
        snap_type="base",
    )

    for build_item in build_plan:
        with check():
            assert build_item.build_base == expected_base
        with check():
            assert (build_item.build_on, build_item.build_for) in expected_archs[
                build_item.platform
            ]


@pytest.mark.parametrize(
    ("base", "build_base", "error_cls", "error_match"),
    [
        (
            "core24",
            "ubuntu@24.04",
            craft_platforms.InvalidBaseError,
            "non-kernel snaps cannot use 'base: coreXY' and arbitrary build-bases",
        ),
        (
            None,
            "devel",
            craft_platforms.RequiresBaseError,
            "snaps of type None require a 'base'",
        ),
    ],
)
def test_get_platforms_snap_build_plan_error(
    base,
    build_base,
    error_cls,
    error_match,
):
    with pytest.raises(error_cls, match=error_match):
        _build.get_platforms_snap_build_plan(
            base=base,
            build_base=build_base,
            platforms=None,
        )


@pytest.mark.parametrize(
    ("base", "build_base", "platforms", "expected"),
    [
        pytest.param(
            "core22",
            None,
            {"amd64": None},
            [
                craft_platforms.BuildInfo(
                    "amd64",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DistroBase("ubuntu", "22.04"),
                ),
            ],
            id="jammy-amd64",
        ),
        pytest.param(
            "core24",
            "devel",
            {"amd64": None},
            [
                craft_platforms.BuildInfo(
                    "amd64",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DistroBase("ubuntu", "devel"),
                ),
            ],
            id="devel-for-noble",
        ),
        pytest.param(
            "core24",
            None,
            {
                "my-desktop": {
                    "build-on": ["amd64"],
                    "build-for": ["amd64"],
                },
                "raspi": {"build-on": ["amd64", "arm64"], "build-for": ["arm64"]},
                "some-mainframe-cross-compile": {
                    "build-on": ["amd64", "arm64"],
                    "build-for": ["s390x"],
                },
            },
            [
                craft_platforms.BuildInfo(
                    "my-desktop",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
                craft_platforms.BuildInfo(
                    "raspi",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.ARM64,
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
                craft_platforms.BuildInfo(
                    "raspi",
                    craft_platforms.DebianArchitecture.ARM64,
                    craft_platforms.DebianArchitecture.ARM64,
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
                craft_platforms.BuildInfo(
                    "some-mainframe-cross-compile",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.S390X,
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
                craft_platforms.BuildInfo(
                    "some-mainframe-cross-compile",
                    craft_platforms.DebianArchitecture.ARM64,
                    craft_platforms.DebianArchitecture.S390X,
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
            ],
            id="multiple-builds",
        ),
    ],
)
def test_build_plans_in_depth(base, build_base, platforms, expected):
    """Test the exact build plan for a set of items."""
    actual = _build.get_platforms_snap_build_plan(
        base=base,
        build_base=build_base,
        platforms=platforms,
    )

    assert actual == expected


@pytest.mark.parametrize(
    ("base", "build_base", "expected_archs"),
    [
        ("core", None, _build.CORE16_18_DEFAULT_ARCHITECTURES),
        ("core16", None, _build.CORE16_18_DEFAULT_ARCHITECTURES),
        ("core18", None, _build.CORE16_18_DEFAULT_ARCHITECTURES),
        ("core20", None, _build.CORE20_DEFAULT_ARCHITECTURES),
        ("core22", None, _build.DEFAULT_ARCHITECTURES),
        ("core24", None, _build.DEFAULT_ARCHITECTURES),
    ],
)
def test_build_plans_default_architectures(base, build_base, expected_archs):
    actual = _build.get_platforms_snap_build_plan(
        base=base,
        build_base=build_base,
        platforms=None,
    )
    actual_archs = [item.build_for for item in actual]
    pytest_check.equal(actual_archs, list(expected_archs))
    for info in actual:
        pytest_check.equal(info.build_on, info.build_for)
        pytest_check.is_in(info.build_for, expected_archs)



# This file is part of craft-platforms.
#
# Copyright 2024 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Package base for craft_platforms."""

from ._architectures import DebianArchitecture, parse_base_and_architecture
from ._build import get_build_plan
from ._buildinfo import BuildInfo
from . import charm, rock, snap
from ._distro import BaseName, DistroBase, is_ubuntu_like
from ._errors import (
    CraftError,
    CraftPlatformsError,
    AllOnlyBuildError,
    AllSinglePlatformError,
    InvalidBaseError,
    InvalidPlatformNameError,
    InvalidPlatformError,
    NeedBuildBaseError,
    RequiresBaseError,
    InvalidMultiBaseError,
)
from ._platforms import (
    PlatformDict,
    Platforms,
    get_platforms_build_plan,
    parse_base_and_name,
)

try:
    from ._version import (
        __version__,
    )  # pyright: ignore[reportMissingImports,reportUnknownVariableType]
except ImportError:  # pragma: no cover
    from importlib.metadata import version, PackageNotFoundError

    try:
        __version__ = version("craft-platforms")
    except PackageNotFoundError:
        __version__ = "dev"

__all__ = [
    "__version__",
    "DebianArchitecture",
    "get_build_plan",
    "BuildInfo",
    "parse_base_and_architecture",
    "charm",
    "rock",
    "snap",
    "get_platforms_build_plan",
    "parse_base_and_name",
    "PlatformDict",
    "Platforms",
    "BaseName",
    "DistroBase",
    "is_ubuntu_like",
    "CraftError",
    "CraftPlatformsError",
    "AllOnlyBuildError",
    "AllSinglePlatformError",
    "InvalidBaseError",
    "InvalidPlatformNameError",
    "InvalidPlatformError",
    "RequiresBaseError",
    "NeedBuildBaseError",
    "InvalidMultiBaseError",
]



# This file is part of craft-platforms.
#
# Copyright 2024 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for distro utilities."""

import itertools

import craft_platforms
import distro
import pytest
import pytest_check
from craft_platforms.test import strategies
from hypothesis import given
from hypothesis import strategies as hp_strat

CENTOS_7 = """\
NAME="CentOS Linux"
VERSION="7 (Core)"
ID="centos"
ID_LIKE="rhel fedora"
VERSION_ID="7"
PRETTY_NAME="CentOS Linux 7 (Core)"
ANSI_COLOR="0;31"
CPE_NAME="cpe:/o:centos:centos:7"
HOME_URL="https://www.centos.org/"
BUG_REPORT_URL="https://bugs.centos.org/"

CENTOS_MANTISBT_PROJECT="CentOS-7"
CENTOS_MANTISBT_PROJECT_VERSION="7"
REDHAT_SUPPORT_PRODUCT="centos"
REDHAT_SUPPORT_PRODUCT_VERSION="7"
"""
DEBIAN_10 = """\
PRETTY_NAME="Debian GNU/Linux 10 (buster)"
NAME="Debian GNU/Linux"
VERSION_ID="10"
VERSION="10 (buster)"
VERSION_CODENAME=buster
ID=debian
HOME_URL="https://www.debian.org/"
SUPPORT_URL="https://www.debian.org/support"
BUG_REPORT_URL="https://bugs.debian.org/"
"""

# Sample linux distributions used for tests. These don't need to be supported.
DAPPER = craft_platforms.DistroBase("ubuntu", "6.06")
BIONIC = craft_platforms.DistroBase("ubuntu", "18.04")
FOCAL = craft_platforms.DistroBase("ubuntu", "20.04")
JAMMY = craft_platforms.DistroBase("ubuntu", "22.04")
NOBLE = craft_platforms.DistroBase("ubuntu", "24.04")
ORACULAR = craft_platforms.DistroBase("ubuntu", "24.10")
DEVEL = craft_platforms.DistroBase("ubuntu", "devel")
BUSTER = craft_platforms.DistroBase("debian", "10")
BOOKWORM = craft_platforms.DistroBase("debian", "12")
ALMA_EIGHT = craft_platforms.DistroBase("almalinux", "8.10")
ALMA_NINE = craft_platforms.DistroBase("almalinux", "9")
ALMA_NINE_FOUR = craft_platforms.DistroBase("almalinux", "9.4")

ALL_UBUNTU = [DAPPER, BIONIC, FOCAL, JAMMY, NOBLE, ORACULAR, DEVEL]
ALL_DEBIAN = [BUSTER, BOOKWORM]
ALL_ALMA = [ALMA_EIGHT, ALMA_NINE]
ALL_DISTROS = [*ALL_UBUNTU, *ALL_DEBIAN, *ALL_ALMA]


@pytest.mark.parametrize(
    ("base", "other", "expected"),
    [
        *[(distro, distro, True) for distro in ALL_DISTROS],
        *[
            (first, second, False)
            for first, second in itertools.permutations(ALL_DISTROS, r=2)
        ],
        *[(distro, ("bsd", "4.2.2"), False) for distro in ALL_DISTROS],
        *[
            pytest.param(
                distro,
                (distro.distribution, distro.series),
                True,
                id=f"{distro.series}_tuple",
            )
            for distro in ALL_DISTROS
        ],
        (ALMA_NINE, ALMA_NINE_FOUR, True),
    ],
)
def test_distro_base_equality(base, other, expected):
    assert (base == other) == expected
    assert (other == base) == expected  # Ensure equality works both ways


@pytest.mark.parametrize(
    ("smaller", "bigger"),
    [
        *[
            pytest.param(
                craft_platforms.DistroBase("ubuntu", "4.10"),
                version,
                id=f"ubuntu_warty_vs_{version.series}",
            )
            for version in ALL_UBUNTU
        ],
        *[
            pytest.param(
                version,
                craft_platforms.DistroBase("ubuntu", "999999.10"),
                id=f"ubuntu_infinity_vs_{version.series}",
            )
            # ubuntu 999999.10 is greater than all ubuntu releases except ubuntu@devel
            for version in ALL_UBUNTU
            if version != DEVEL
        ],
        *[
            pytest.param(
                version,
                craft_platforms.DistroBase("ubuntu", "devel"),
                id=f"ubuntu_devel_vs_{version.series}",
            )
            # ubuntu@devel is greater than all ubuntu releases
            for version in ALL_UBUNTU
            if version != DEVEL
        ],
        *[
            pytest.param(
                craft_platforms.DistroBase("debian", "1.1"),
                version,
                id=f"debian_buzz_vs_{version.series}",
            )
            for version in ALL_DEBIAN
        ],
        *[
            pytest.param(
                version,
                craft_platforms.DistroBase("debian", "99999"),
                id=f"debian_future_vs_{version.series}",
            )
            for version in ALL_DEBIAN
        ],
        *[
            pytest.param(
                craft_platforms.DistroBase("almalinux", "0"),
                version,
                id=f"alma_zero_vs_{version.series}",
            )
            for version in ALL_ALMA
        ],
        *[
            pytest.param(
                version,
                craft_platforms.DistroBase("almalinux", "99999"),
                id=f"alma_future_vs_{version.series}",
            )
            for version in ALL_ALMA
        ],
    ],
)
def test_distro_base_difference_success(check, smaller, bigger):
    with check():
        assert bigger > smaller
    with check():
        assert bigger >= smaller
    with check():
        assert smaller <= bigger
    with check():
        assert smaller < bigger
    with check():
        assert bigger != smaller


@pytest.mark.parametrize(
    ("first", "second"),
    [
        *list(itertools.product(ALL_UBUNTU, ALL_DEBIAN)),
        *list(itertools.product(ALL_UBUNTU, ALL_ALMA)),
        *list(itertools.product(ALL_DEBIAN, ALL_ALMA)),
    ],
)
def test_compare_incompatible_distros(check, first, second):
    pytest_check.is_false(first == second)
    pytest_check.is_false(second == first)
    pytest_check.is_true(first != second)
    pytest_check.is_true(second != first)
    with check.raises(ValueError):
        assert first > second
    with check.raises(ValueError):
        assert first >= second
    with check.raises(ValueError):
        assert first <= second
    with check.raises(ValueError):
        assert first < second
    with check.raises(ValueError):
        assert second > first
    with check.raises(ValueError):
        assert second >= first
    with check.raises(ValueError):
        assert second <= first
    with check.raises(ValueError):
        assert second < first


@pytest.mark.parametrize(
    ("first", "second"),
    [
        (NOBLE, ("ubuntu", "24.04", "this should not be here.")),
        (NOBLE, ("ubuntu", 24.04)),
    ],
)
def test_compare_incompatible_types(check, first, second):
    pytest_check.is_false(first == second)
    pytest_check.is_false(second == first)
    pytest_check.is_true(first != second)
    pytest_check.is_true(second != first)
    with check.raises(TypeError):
        assert first > second
    with check.raises(TypeError):
        assert first >= second
    with check.raises(TypeError):
        assert first <= second
    with check.raises(TypeError):
        assert first < second
    with check.raises(TypeError):
        assert second > first
    with check.raises(TypeError):
        assert second >= first
    with check.raises(TypeError):
        assert second <= first
    with check.raises(TypeError):
        assert second < first


@pytest.mark.parametrize(
    ("os_release", "expected"),
    [
        (CENTOS_7, False),
        (DEBIAN_10, False),
    ],
)
def test_is_ubuntu_like(os_release: str, expected):
    distribution = distro.LinuxDistribution(
        include_lsb=False,
        os_release_file=os_release,
    )
    assert craft_platforms.is_ubuntu_like(distribution) is expected


@pytest.mark.parametrize(
    ("distro_string", "expected"),
    [
        ("ubuntu@6.06", DAPPER),
        ("ubuntu@18.04", BIONIC),
        ("ubuntu@20.04", FOCAL),
        ("ubuntu@22.04", JAMMY),
        ("ubuntu@24.04", NOBLE),
        ("ubuntu@24.10", ORACULAR),
        ("devel", DEVEL),
        ("ubuntu@devel", DEVEL),
        ("debian@10", BUSTER),
        ("debian@12", BOOKWORM),
        ("almalinux@8.10", ALMA_EIGHT),
        ("almalinux@9.4", ALMA_NINE),
    ],
)
def test_from_str(distro_string, expected):
    actual = craft_platforms.DistroBase.from_str(distro_string)

    assert actual == expected


def test_from_str_error():
    with pytest.raises(ValueError, match="Invalid base string 'invalid-base'.*"):
        craft_platforms.DistroBase.from_str("invalid-base")


@given(base=strategies.any_distro_base())
def test_fuzz_distrobase_equality(base: craft_platforms.DistroBase):
    assert base == base  # noqa: PLR0124
    assert base >= base  # noqa: PLR0124
    assert base <= base  # noqa: PLR0124
    assert not (base > base)  # noqa: PLR0124
    assert not (base < base)  # noqa: PLR0124


@given(
    distro=hp_strat.text(hp_strat.characters(blacklist_characters="@")),
    series=hp_strat.text(hp_strat.characters(blacklist_characters="@")),
)
def test_fuzz_distrobase(distro: str, series: str):
    base = craft_platforms.DistroBase(distribution=distro, series=series)
    assert base.distribution == distro
    assert base.series == series
    assert str(base) == f"{distro}@{series}"
    assert base == craft_platforms.DistroBase.from_str(f"{distro}@{series}")


@given(hp_strat.builds(distro.LinuxDistribution))
def test_fuzz_distrobase_from_linux_distribution(
    distribution: distro.LinuxDistribution,
):
    craft_platforms.DistroBase.from_linux_distribution(distribution)


@given(hp_strat.builds(distro.LinuxDistribution))
def test_fuzz_is_ubuntu_like(distribution):
    craft_platforms.is_ubuntu_like(distribution)



# This file is part of craft-platforms.
#
# Copyright 2024 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Distribution related utilities."""

from __future__ import annotations

import contextlib
import dataclasses
import typing
from typing import List, Union, cast

import distro
from typing_extensions import Self


@typing.runtime_checkable
class BaseName(typing.Protocol):
    """A protocol for any class that can be used as an OS base.

    This protocol exists as a backwards compatibility shim for the
    language used in craft-providers.
    """

    @property
    def name(self) -> str: ...

    @property
    def version(self) -> str: ...


def _get_series_tuple(version_str: str) -> tuple[Union[int, str], ...]:
    """Convert a version string into a version tuple."""
    parts = typing.cast(List[Union[str, int]], version_str.split("."))
    # Try converting each part to an integer, leaving as a string if not doable.
    for idx, part in enumerate(parts):
        with contextlib.suppress(ValueError):
            parts[idx] = int(part)
    return tuple(parts)


def _get_distro(base: Union[DistroBase, BaseName] | tuple[str, str]) -> str:
    """Get the distribution of a base."""
    if isinstance(base, DistroBase):
        return base.distribution
    if isinstance(base, BaseName):
        return base.name
    return base[0]


def _get_series(base: Union[DistroBase, BaseName] | tuple[str, str]) -> str:
    """Get the version of a base."""
    if isinstance(base, DistroBase):
        return base.series
    if isinstance(base, BaseName):
        return base.version
    return base[1]


def _is_distrobase_compatible(item: object) -> bool:
    """Check if the given object is compatible for comparison with a DistroBase."""
    if isinstance(item, (DistroBase, BaseName)):
        return True
    return (
        isinstance(item, tuple)
        and len(item) == 2  # noqa: PLR2004 (if it's a tuple, it'll be ('distro', 'series'))
        and isinstance(item[0], str)
        and isinstance(item[1], str)
    )


@dataclasses.dataclass(repr=True, frozen=True)
class DistroBase:  # noqa: PLW1641 (https://github.com/astral-sh/ruff/issues/18905)
    """A linux distribution base."""

    distribution: str
    series: str

    def _ensure_bases_comparable(
        self,
        other: Union[DistroBase, BaseName] | tuple[str, str],
    ) -> None:
        """Ensure that these bases are comparable, raising an exception if not.

        :param other: Another distribution base.
        :raises: ValueError if the distribution bases are not comparable.
        """
        other_distro = _get_distro(other)
        if self.distribution != other_distro:
            raise ValueError(
                f"Different distributions ({self.distribution} and {other_distro}) do not have comparable versions.",
            )

    def __eq__(self, other: object, /) -> bool:
        if not _is_distrobase_compatible(other):
            return NotImplemented
        other = cast("DistroBase | BaseName | tuple[str, str]", other)
        other_distro = _get_distro(other)

        if self.distribution != other_distro:
            return False

        other_series = _get_series(other)
        # The series is allowed to be more specific on one side.
        return all(
            this_part == other_part
            for this_part, other_part in zip(
                self.series.split("."), other_series.split(".")
            )
        )

    def __lt__(self, other: object) -> bool:
        if not _is_distrobase_compatible(other):
            return NotImplemented
        other = cast("DistroBase | BaseName | tuple[str, str]", other)
        self._ensure_bases_comparable(other)
        other_version = _get_series(other)
        if self.series == "devel" or other_version == "devel":
            return self.series != "devel" and other_version == "devel"
        self_version_tuple = _get_series_tuple(self.series)
        other_version_tuple = _get_series_tuple(other_version)
        return self_version_tuple < other_version_tuple

    def __le__(self, other: object) -> bool:
        if not _is_distrobase_compatible(other):
            return NotImplemented
        other = cast("DistroBase | BaseName | tuple[str, str]", other)
        self._ensure_bases_comparable(other)
        other_version = _get_series(other)
        if self.series == "devel" or other_version == "devel":
            return other_version == "devel"
        self_version_tuple = _get_series_tuple(self.series)
        other_version_tuple = _get_series_tuple(other_version)
        return self_version_tuple <= other_version_tuple

    def __gt__(self, other: object) -> bool:
        if not _is_distrobase_compatible(other):
            return NotImplemented
        other = cast("DistroBase | BaseName | tuple[str, str]", other)
        self._ensure_bases_comparable(other)
        other_version = _get_series(other)
        if self.series == "devel" or other_version == "devel":
            return other_version != "devel"
        self_version_tuple = _get_series_tuple(self.series)
        other_version_tuple = _get_series_tuple(other_version)
        return self_version_tuple > other_version_tuple

    def __ge__(self, other: object) -> bool:
        if not _is_distrobase_compatible(other):
            return NotImplemented
        other = cast("DistroBase | BaseName | tuple[str, str]", other)
        self._ensure_bases_comparable(other)
        other_version = _get_series(other)
        if self.series == "devel" or other_version == "devel":
            return self.series == "devel"
        self_version_tuple = _get_series_tuple(self.series)
        other_version_tuple = _get_series_tuple(_get_series(other))
        return self_version_tuple >= other_version_tuple

    @classmethod
    def from_str(cls, base_str: str) -> Self:
        """Parse a distribution string to a DistroBase.

        :param base_str: A distribution string (e.g. "ubuntu@24.04")
        :returns: A DistroBase of this string.
        :raises: ValueError if the string isn't of the appropriate format.
        """
        # "devel" is an exception and corresponds to `ubuntu@devel`
        if base_str == "devel":
            return cls("ubuntu", "devel")

        if base_str.count("@") != 1:
            raise ValueError(
                f"Invalid base string {base_str!r}. Format should be '<distribution>@<series>'",
            )
        distribution, _, series = base_str.partition("@")
        return cls(distribution, series)

    @classmethod
    def from_linux_distribution(cls, distribution: distro.LinuxDistribution) -> Self:
        """Convert a distro package's LinuxDistribution object to a DistroBase.

        :param distribution: A LinuxDistribution from the distro package.
        :returns: A matching DistroBase object.
        """
        return cls(distribution=distribution.id(), series=distribution.version())

    def __str__(self) -> str:
        return f"{self.distribution}@{self.series}"


def is_ubuntu_like(distribution: Union[distro.LinuxDistribution, None] = None) -> bool:
    """Determine whether the given distribution is Ubuntu or Ubuntu-like.

    :param distribution: Linux distribution info object, or None to use the host system.
    :returns: A boolean noting whether the given distribution is Ubuntu or Ubuntu-like.
    """
    if distribution is None:
        distribution = distro.LinuxDistribution()
    if distribution.id() == "ubuntu":
        return True
    distros_like = distribution.like().split()
    return "ubuntu" in distros_like



# This file is part of craft-platforms.
#
# Copyright 2024 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Error classes for craft-platforms."""

import dataclasses
import os
import typing
from typing import Collection, Iterable, Optional

# Workaround for Windows...
EX_DATAERR = getattr(os, "EX_DATAERR", 65)


@typing.runtime_checkable
class CraftError(typing.Protocol):
    """A protocol for determining whether an object is a craft error."""

    args: typing.Collection[str]
    details: Optional[str]
    resolution: Optional[str]


@dataclasses.dataclass(unsafe_hash=True)
class CraftPlatformsError(Exception):  # noqa: PLW1641 (https://github.com/astral-sh/ruff/issues/18905)
    """Signal a program error with a lot of information to report."""

    message: str
    """The main message to the user about this error."""

    details: Optional[str] = None
    """The full error details which originated the error situation."""

    resolution: Optional[str] = None
    """An extra line indicating to the user how the error may be fixed or avoided (to be
      shown together with ``message``)."""

    docs_url: Optional[str] = None
    """An URL to point the user to documentation (to be shown together with ``message``)."""

    doc_slug: Optional[str] = None
    """The slug to the user documentation. Needs a base url to form a full address.
      Note that ``docs_url`` has preference if it is set."""

    logpath_report: bool = True
    """Whether the location of the log filepath should be presented in the screen as the
     final message."""

    reportable: bool = True
    """If an error report should be sent to some error-handling backend (like Sentry)."""

    retcode: int = 1
    """The code to return when the application finishes."""

    def __post_init__(self) -> None:
        super().__init__(self.message)
        if self.doc_slug and not self.doc_slug.startswith("/"):
            self.doc_slug = f"/{self.doc_slug}"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, CraftPlatformsError):
            return (
                self.message == other.message
                and self.details == other.details
                and self.resolution == other.resolution
                and self.docs_url == other.docs_url
                and self.logpath_report == other.logpath_report
                and self.reportable == other.reportable
                and self.retcode == other.retcode
                and self.doc_slug == other.doc_slug
            )
        if isinstance(other, CraftError) and isinstance(other, Exception):
            if (
                self.args != other.args
                or self.details != other.details
                or self.resolution != other.resolution
            ):
                return False
            for attr in (
                "message",
                "docs_url",
                "docs_slug",
                "logpath_report",
                "reportable",
                "retcode",
            ):
                if hasattr(other, attr) and getattr(other, attr) != getattr(self, attr):
                    return False
            return True
        return NotImplemented


class BuildForAllError(CraftPlatformsError, ValueError):
    """Errors related to build-for: all."""


class AllOnlyBuildError(BuildForAllError):
    """Error when multiple build-for architectures are defined, but one is 'all'."""

    def __init__(
        self,
        platforms: Iterable[str],
    ) -> None:
        bfa_platforms = ",".join(platforms)
        super().__init__(
            message="build-for: all must be the only build-for architecture",
            details=f"build-for: all defined in platforms: {bfa_platforms}",
            resolution="Provide only one platform with only build-for: all or remove 'all' from build-for options.",
        )


class AllSinglePlatformError(BuildForAllError):
    """Error when multiple build-for architectures are defined, but one is 'all'."""

    def __init__(
        self,
        platforms: Collection[str],
    ) -> None:
        bfa_platforms = ",".join(platforms)
        super().__init__(
            message=f"build-for: all requires exactly one platform definition ({len(platforms)} provided)",
            details=f"build-for: all defined in platforms: {bfa_platforms}",
            resolution="Provide only one platform with only build-for: all or remove 'all' from build-for options.",
        )


class NeedBuildBaseError(CraftPlatformsError, ValueError):
    """Error when ``base`` requires a ``build_base``, but none is unspecified."""

    def __init__(self, base: str) -> None:
        super().__init__(
            message=f"base '{base}' requires a 'build-base', but none is specified",
            resolution="Specify a build-base.",
            retcode=EX_DATAERR,
        )


class InvalidPlatformNameError(CraftPlatformsError, ValueError):
    """Error when a specified platform name is not a Debian architecture."""

    def __init__(self, platform_name: str) -> None:
        self.platform_name = platform_name
        super().__init__(
            message=f"platform name {platform_name!r} is not a valid Debian architecture and needs 'build-on' and 'build-for' specified",
            resolution=f"Specify 'build-on' and 'build-for' values under the {platform_name!r} entry.",
        )


class InvalidPlatformError(CraftPlatformsError, ValueError):
    """Error when a specified platform is invalid."""

    def __init__(
        self,
        platform_name: str,
        *,
        details: Optional[str] = None,
        resolution: str,
        docs_url: Optional[str] = None,
        doc_slug: Optional[str] = None,
    ) -> None:
        self.platform_name = platform_name
        super().__init__(
            message=f"platform {platform_name!r} is invalid",
            details=details,
            resolution=resolution,
            docs_url=docs_url,
            doc_slug=doc_slug,
        )


class InvalidBaseError(CraftPlatformsError, ValueError):
    """Error when a specified base name is invalid."""

    def __init__(
        self,
        base: str,
        *,
        message: Optional[str] = None,
        resolution: Optional[str] = None,
        docs_url: Optional[str] = None,
        build_base: bool = False,
    ) -> None:
        self.base = base
        if resolution is None:
            resolution = "Ensure the base matches the <distro>@<series> pattern and is a supported series."
        if not message:
            message = (
                f"build-base '{base}' is unknown or invalid"
                if build_base
                else f"base '{base}' is unknown or invalid"
            )

        super().__init__(message=message, resolution=resolution, docs_url=docs_url)


class RequiresBaseError(CraftPlatformsError, ValueError):
    """Error when a base is required in this configuration."""


class InvalidMultiBaseError(CraftPlatformsError, ValueError):
    """Error when a multi-base configuration is invalid."""



# This file is part of craft-platforms.
#
# Copyright 2024 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for charmcraft builds."""

import itertools
from typing import Optional

import craft_platforms
import pytest
import pytest_check
from craft_platforms import charm
from craft_platforms.test import strategies
from hypothesis import assume, given
from hypothesis import strategies as hp_strat

SAMPLE_UBUNTU_VERSIONS = ("16.04", "18.04", "20.04", "22.04", "24.04", "24.10", "devel")


@pytest.mark.parametrize(
    ("base", "build_base", "expected_base"),
    [
        *[
            # No special build base
            (f"ubuntu@{version}", None, craft_platforms.DistroBase("ubuntu", version))
            for version in SAMPLE_UBUNTU_VERSIONS
        ],
        *[
            # Always build on a different Ubuntu version
            (
                "ubuntu@00.04",
                f"ubuntu@{version}",
                craft_platforms.DistroBase("ubuntu", version),
            )
            for version in SAMPLE_UBUNTU_VERSIONS
        ],
    ],
)
@pytest.mark.parametrize(
    ("platforms", "platform_archs"),
    [
        pytest.param(
            None,
            {
                architecture.value: [(architecture, architecture)]
                for architecture in charm.DEFAULT_ARCHITECTURES
            },
            id="default-platforms",
        ),
        *[
            pytest.param(
                {architecture.value: None},
                {architecture.value: [(architecture.value, architecture.value)]},
                id=f"implicit-{architecture.value}",
            )
            for architecture in craft_platforms.DebianArchitecture
        ],
        *[
            pytest.param(
                {
                    architecture.value: {
                        "build-on": [architecture.value],
                        "build-for": [architecture.value],
                    },
                },
                {architecture.value: [(architecture.value, architecture.value)]},
                id=f"explicit-{architecture.value}",
            )
            for architecture in craft_platforms.DebianArchitecture
        ],
        *[
            pytest.param(
                {
                    architecture.value: {
                        "build-on": architecture.value,
                        "build-for": architecture.value,
                    },
                },
                {architecture.value: [(architecture.value, architecture.value)]},
                id=f"explicit-scalar-{architecture.value}",
            )
            for architecture in craft_platforms.DebianArchitecture
        ],
        *[
            pytest.param(
                {
                    "my-platform": {
                        "build-on": [
                            arch.value for arch in craft_platforms.DebianArchitecture
                        ],
                        "build-for": [build_for_arch.value],
                    },
                },
                {
                    "my-platform": [
                        (arch.value, build_for_arch.value)
                        for arch in craft_platforms.DebianArchitecture
                    ],
                },
                id=f"build-on-any-for-{build_for_arch.value}",
            )
            for build_for_arch in craft_platforms.DebianArchitecture
        ],
        *[
            pytest.param(
                {
                    "my-platform": {
                        "build-on": [
                            arch.value for arch in craft_platforms.DebianArchitecture
                        ],
                        "build-for": build_for_arch.value,
                    },
                },
                {
                    "my-platform": [
                        (arch.value, build_for_arch.value)
                        for arch in craft_platforms.DebianArchitecture
                    ],
                },
                id=f"build-on-any-for-scalar-{build_for_arch.value}",
            )
            for build_for_arch in craft_platforms.DebianArchitecture
        ],
    ],
)
def test_build_plans_success(
    check,
    base,
    build_base,
    expected_base,
    platforms,
    platform_archs,
):
    """Shallow test for success on a large number of platform items."""
    build_plan = charm.get_platforms_charm_build_plan(
        base=base,
        build_base=build_base,
        platforms=platforms,
    )

    for build_item in build_plan:
        with check():
            assert build_item.build_base == expected_base
        with check():
            assert (build_item.build_on, build_item.build_for) in platform_archs[
                build_item.platform
            ]
    expected_length = len(
        list(
            itertools.chain.from_iterable(
                arch_pairs for arch_pairs in platform_archs.values()
            ),
        ),
    )
    pytest_check.equal(expected_length, len(build_plan))


@pytest.mark.parametrize(
    ("base", "build_base", "platforms", "expected"),
    [
        pytest.param(
            "ubuntu@22.04",
            None,
            {"amd64": None},
            [
                craft_platforms.BuildInfo(
                    "amd64",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DistroBase("ubuntu", "22.04"),
                ),
            ],
            id="jammy-amd64",
        ),
        pytest.param(
            "ubuntu@24.04",
            "ubuntu@22.04",
            {"amd64": None},
            [
                craft_platforms.BuildInfo(
                    "amd64",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DistroBase("ubuntu", "22.04"),
                ),
            ],
            id="jammy-for-noble-amd64",
        ),
        pytest.param(
            "ubuntu@24.04",
            None,
            {
                "my-desktop": {
                    "build-on": ["amd64"],
                    "build-for": ["amd64"],
                },
                "raspi": {"build-on": ["amd64", "arm64"], "build-for": ["arm64"]},
                "some-mainframe-cross-compile": {
                    "build-on": ["amd64", "arm64"],
                    "build-for": ["s390x"],
                },
            },
            [
                craft_platforms.BuildInfo(
                    "my-desktop",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
                craft_platforms.BuildInfo(
                    "raspi",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.ARM64,
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
                craft_platforms.BuildInfo(
                    "raspi",
                    craft_platforms.DebianArchitecture.ARM64,
                    craft_platforms.DebianArchitecture.ARM64,
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
                craft_platforms.BuildInfo(
                    "some-mainframe-cross-compile",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.S390X,
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
                craft_platforms.BuildInfo(
                    "some-mainframe-cross-compile",
                    craft_platforms.DebianArchitecture.ARM64,
                    craft_platforms.DebianArchitecture.S390X,
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
            ],
            id="multiple-builds",
        ),
        pytest.param(
            None,
            None,
            {
                "noble": {
                    "build-on": ["ubuntu@24.04:amd64"],
                    "build-for": ["ubuntu@24.04:amd64"],
                },
            },
            [
                craft_platforms.BuildInfo(
                    "noble",
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                )
            ],
            id="multi-base-simple",
        ),
        pytest.param(
            None,
            None,
            {
                "noble": {
                    "build-on": "ubuntu@24.04:amd64",
                    "build-for": "ubuntu@24.04:amd64",
                },
            },
            [
                craft_platforms.BuildInfo(
                    "noble",
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                )
            ],
            id="multi-base-scalar",
        ),
        pytest.param(
            None,
            None,
            {"ubuntu@24.04:amd64": None},
            [
                craft_platforms.BuildInfo(
                    "ubuntu@24.04:amd64",
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                )
            ],
            id="multi-base-shorthand",
        ),
        pytest.param(
            None,
            None,
            {
                # base and arch in platform name
                "ubuntu@20.04:amd64": None,
                # base and arch in build entries
                "noble": {
                    "build-on": ["ubuntu@24.04:amd64"],
                    "build-for": ["ubuntu@24.04:amd64"],
                },
            },
            [
                craft_platforms.BuildInfo(
                    "ubuntu@20.04:amd64",
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DistroBase("ubuntu", "20.04"),
                ),
                craft_platforms.BuildInfo(
                    "noble",
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
            ],
            id="multi-base-mixed-notation",
        ),
        pytest.param(
            None,
            None,
            {
                "jammy": {
                    "build-on": ["ubuntu@22.04:amd64"],
                    "build-for": ["ubuntu@22.04:all"],
                },
                "noble": {
                    "build-on": ["ubuntu@24.04:amd64"],
                    "build-for": ["ubuntu@24.04:all"],
                },
            },
            [
                craft_platforms.BuildInfo(
                    "jammy",
                    craft_platforms.DebianArchitecture("amd64"),
                    "all",
                    craft_platforms.DistroBase("ubuntu", "22.04"),
                ),
                craft_platforms.BuildInfo(
                    "noble",
                    craft_platforms.DebianArchitecture("amd64"),
                    "all",
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
            ],
            id="multi-base-all",
        ),
        pytest.param(
            None,
            None,
            {
                "jammy": {
                    "build-on": "ubuntu@22.04:amd64",
                    "build-for": "ubuntu@22.04:all",
                },
                "noble": {
                    "build-on": "ubuntu@24.04:amd64",
                    "build-for": "ubuntu@24.04:all",
                },
            },
            [
                craft_platforms.BuildInfo(
                    "jammy",
                    craft_platforms.DebianArchitecture("amd64"),
                    "all",
                    craft_platforms.DistroBase("ubuntu", "22.04"),
                ),
                craft_platforms.BuildInfo(
                    "noble",
                    craft_platforms.DebianArchitecture("amd64"),
                    "all",
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
            ],
            id="multi-base-all-scalar",
        ),
        pytest.param(
            None,
            None,
            {
                "ubuntu@20.04:amd64": None,
                "jammy-list": {
                    "build-on": ["ubuntu@22.04:amd64"],
                    "build-for": ["ubuntu@22.04:amd64"],
                },
                "jammy-scalar": {
                    "build-on": "ubuntu@22.04:amd64",
                    "build-for": "ubuntu@22.04:amd64",
                },
                "noble": {
                    "build-on": ["ubuntu@22.04:amd64"],
                    "build-for": ["ubuntu@22.04:amd64"],
                },
                "noble-cross": {
                    "build-on": ["ubuntu@24.04:amd64", "ubuntu@24.04:riscv64"],
                    "build-for": ["ubuntu@24.04:riscv64"],
                },
            },
            [
                craft_platforms.BuildInfo(
                    "ubuntu@20.04:amd64",
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DistroBase("ubuntu", "20.04"),
                ),
                craft_platforms.BuildInfo(
                    "jammy-list",
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DistroBase("ubuntu", "22.04"),
                ),
                craft_platforms.BuildInfo(
                    "jammy-scalar",
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DistroBase("ubuntu", "22.04"),
                ),
                craft_platforms.BuildInfo(
                    "noble",
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DistroBase("ubuntu", "22.04"),
                ),
                craft_platforms.BuildInfo(
                    "noble-cross",
                    craft_platforms.DebianArchitecture("amd64"),
                    craft_platforms.DebianArchitecture("riscv64"),
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
                craft_platforms.BuildInfo(
                    "noble-cross",
                    craft_platforms.DebianArchitecture("riscv64"),
                    craft_platforms.DebianArchitecture("riscv64"),
                    craft_platforms.DistroBase("ubuntu", "24.04"),
                ),
            ],
            id="multi-base-complex",
        ),
    ],
)
def test_build_plans_in_depth(base, build_base, platforms, expected):
    """Test the exact build plan for a set of items."""
    actual = charm.get_platforms_charm_build_plan(
        base=base,
        build_base=build_base,
        platforms=platforms,
    )

    assert actual == expected


@pytest.mark.parametrize(
    ("base", "build_base", "platforms", "error_msg", "error_res"),
    [
        pytest.param(
            "invalid-base",
            None,
            None,
            "Invalid base string 'invalid-base'. Format should be '<distribution>@<series>'",
            None,
            id="invalid-base",
        ),
        pytest.param(
            None,
            None,
            None,
            "No base, build-base, or platforms are declared.",
            "Declare a base or build-base.",
            id="no-base-no-platform",
        ),
        pytest.param(
            None,
            None,
            {
                "my-platform": {
                    "build-on": ["amd64"],
                    "build-for": ["amd64"],
                },
            },
            "No base or build-base is declared and no base is declared in the platforms section.",
            "Declare a base or build-base.",
            id="no-base-with-platform",
        ),
        pytest.param(
            None,
            None,
            {"amd64": None},
            "No base or build-base is declared and no base is declared in the platforms section.",
            "Declare a base or build-base.",
            id="no-base-with-shorthand-platform",
        ),
        pytest.param(
            "ubuntu@24.04",
            None,
            {"ubuntu@24.04:amd64": None},
            "Platform 'ubuntu@24.04:amd64' declares a base and a top-level base or build-base is declared.",
            "Remove the base from the platform's name or remove the top-level base or build-base.",
            id="base-and-platform-base",
        ),
        pytest.param(
            None,
            "ubuntu@24.04",
            {"ubuntu@24.04:amd64": None},
            "Platform 'ubuntu@24.04:amd64' declares a base and a top-level base or build-base is declared.",
            "Remove the base from the platform's name or remove the top-level base or build-base.",
            id="build-base-and-platform-base",
        ),
        pytest.param(
            "ubuntu@24.04",
            None,
            {
                "my-platform": {
                    "build-on": ["ubuntu@24.04:amd64"],
                    "build-for": ["ubuntu@24.04:amd64"],
                },
            },
            "Platform 'my-platform' declares a base and a top-level base or build-base is declared.",
            "Remove the base from the platform's name or remove the top-level base or build-base.",
            id="base-and-build-on-for-base",
        ),
        pytest.param(
            None,
            "ubuntu@24.04",
            {
                "my-platform": {
                    "build-on": ["ubuntu@24.04:amd64"],
                    "build-for": ["ubuntu@24.04:amd64"],
                },
            },
            "Platform 'my-platform' declares a base and a top-level base or build-base is declared.",
            "Remove the base from the platform's name or remove the top-level base or build-base.",
            id="build-base-and-build-on-for-base",
        ),
        pytest.param(
            None,
            None,
            {
                "my-platform": {
                    "build-on": ["ubuntu@22.04:amd64"],
                    "build-for": ["ubuntu@24.04:amd64"],
                },
            },
            "Platform 'my-platform' has mismatched bases in the 'build-on' and 'build-for' entries.",
            "Use the same base for all 'build-on' and 'build-for' entries for the platform.",
            id="build-on-for-base-mismatch",
        ),
        pytest.param(
            None,
            None,
            {
                "my-platform": {
                    "build-on": ["ubuntu@22.04:amd64"],
                    "build-for": ["amd64"],
                },
            },
            "Platform 'my-platform' has mismatched bases in the 'build-on' and 'build-for' entries.",
            "Use the same base for all 'build-on' and 'build-for' entries for the platform.",
            id="build-on-for-base-missing",
        ),
        pytest.param(
            None,
            None,
            {
                "ubuntu@24.04:amd64": {
                    "build-on": ["amd64"],
                    "build-for": ["amd64"],
                },
            },
            "Platform 'ubuntu@24.04:amd64' declares a base in the platform's name and declares 'build-on' and 'build-for' entries.",
            "Either remove the base from the platform's name or remove the 'build-on' and 'build-for' entries for the platform.",
            id="platform-base-with-entries",
        ),
    ],
)
def test_build_plans_bad_base(base, build_base, platforms, error_msg, error_res):
    with pytest.raises(
        (ValueError, craft_platforms.CraftPlatformsError), match=error_msg
    ) as err:
        charm.get_platforms_charm_build_plan(
            base=base,
            build_base=build_base,
            platforms=platforms,
        )

    if error_res and isinstance(err.value, craft_platforms.CraftPlatformsError):
        assert err.value.resolution == error_res


@pytest.mark.parametrize(
    ("platforms", "error_msg"),
    [
        pytest.param(
            {"my machine": None},
            "Platform name 'my machine' is not a valid Debian architecture. Specify a build-on and build-for.",
            id="invalid-platform-name-no-details",
        ),
        pytest.param(
            {"my machine": {"build-on": ["my machine"], "build-for": ["amd64"]}},
            "'my machine' is not a valid Debian architecture",
            id="invalid-architecture-name",
        ),
        pytest.param(
            {"my machine": {"build-on": ["all"], "build-for": ["amd64"]}},
            "Platform 'my machine' has an invalid 'build-on' entry of 'all'.",
            id="build-on-all",
        ),
    ],
)
def test_build_plans_bad_architecture(platforms, error_msg):
    with pytest.raises(ValueError, match=error_msg):
        charm.get_platforms_charm_build_plan("ubuntu@24.04", platforms)


def _is_valid_platform(platforms):
    """Allow test suite to pass until #116 is fixed."""
    for platform in platforms:
        if ":" in platform:
            platform_name, _, platform_base = platform.partition(":")
            if not platform_base or not platform_name or "@" not in platform_name:
                print(f"Skipping unhandled platform name {platform} (#116)")
                return False

    return True


@given(
    base=strategies.real_distro_base(),
    platforms=strategies.platform(
        distro_base=strategies.any_distro_base(),
        shorthand_keys=strategies.build_on_arch_str(),
        values=strategies.platform_dict(
            build_ons=strategies.build_on_arch_str(),
            build_fors=strategies.build_for_arch_str(),
        ),
    ),
    build_base=hp_strat.one_of(hp_strat.none(), strategies.any_distro_base()),
)
def test_fuzz_get_platforms_build_plan_single_base(
    base: craft_platforms.DistroBase,
    platforms: craft_platforms.Platforms,
    build_base: Optional[craft_platforms.DistroBase],
):
    assume(_is_valid_platform(platforms))
    craft_platforms.charm.get_platforms_charm_build_plan(
        base=str(base),
        platforms=platforms,
        build_base=str(build_base) if build_base else None,
    )


@pytest.mark.slow
@given(
    platforms=strategies.platform(
        distro_base=strategies.real_distro_base(),
        shorthand_keys=strategies.distro_series_arch_str(strategies.any_distro_base()),
        values=strategies.platform_dict(
            build_ons=strategies.distro_series_arch_str(strategies.any_distro_base()),
            build_fors=strategies.distro_series_arch_str(strategies.any_distro_base()),
        ).filter(
            lambda p: {p["build-for"][0].partition(":")[0]}
            == {on.partition(":")[0] for on in p["build-on"]}
        ),
    ),
)
def test_fuzz_get_platforms_build_plan_multi_base(
    platforms: craft_platforms.Platforms,
):
    assume(_is_valid_platform(platforms))
    craft_platforms.charm.get_platforms_charm_build_plan(None, platforms)


@pytest.mark.parametrize(
    ("bases", "expected"),
    [
        pytest.param(
            [{"name": "ubuntu", "channel": "20.04"}],
            [
                craft_platforms.BuildInfo(
                    "ubuntu-20.04-amd64",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DistroBase("ubuntu", "20.04"),
                ),
                craft_platforms.BuildInfo(
                    "ubuntu-20.04-arm64",
                    craft_platforms.DebianArchitecture.ARM64,
                    craft_platforms.DebianArchitecture.ARM64,
                    craft_platforms.DistroBase("ubuntu", "20.04"),
                ),
                craft_platforms.BuildInfo(
                    "ubuntu-20.04-ppc64el",
                    craft_platforms.DebianArchitecture.PPC64EL,
                    craft_platforms.DebianArchitecture.PPC64EL,
                    craft_platforms.DistroBase("ubuntu", "20.04"),
                ),
                craft_platforms.BuildInfo(
                    "ubuntu-20.04-riscv64",
                    craft_platforms.DebianArchitecture.RISCV64,
                    craft_platforms.DebianArchitecture.RISCV64,
                    craft_platforms.DistroBase("ubuntu", "20.04"),
                ),
                craft_platforms.BuildInfo(
                    "ubuntu-20.04-s390x",
                    craft_platforms.DebianArchitecture.S390X,
                    craft_platforms.DebianArchitecture.S390X,
                    craft_platforms.DistroBase("ubuntu", "20.04"),
                ),
            ],
        ),
        pytest.param(
            [{"name": "ubuntu", "channel": "22.04", "architectures": ["arm64"]}],
            [
                craft_platforms.BuildInfo(
                    "ubuntu-22.04-arm64",
                    craft_platforms.DebianArchitecture.ARM64,
                    craft_platforms.DebianArchitecture.ARM64,
                    craft_platforms.DistroBase("ubuntu", "22.04"),
                ),
            ],
        ),
        pytest.param(
            [
                {
                    "build-on": [
                        {
                            "name": "ubuntu",
                            "channel": "22.04",
                            "architectures": ["amd64", "s390x"],
                        }
                    ],
                    "run-on": [
                        {
                            "name": "ubuntu",
                            "channel": "22.04",
                            "architectures": ["s390x"],
                        }
                    ],
                },
            ],
            [
                craft_platforms.BuildInfo(
                    "ubuntu-22.04-s390x",
                    craft_platforms.DebianArchitecture.AMD64,
                    craft_platforms.DebianArchitecture.S390X,
                    craft_platforms.DistroBase("ubuntu", "22.04"),
                ),
                craft_platforms.BuildInfo(
                    "ubuntu-22.04-s390x",
                    craft_platforms.DebianArchitecture.S390X,
                    craft_platforms.DebianArchitecture.S390X,
                    craft_platforms.DistroBase("ubuntu", "22.04"),
                ),
            ],
        ),
        pytest.param(
            [
                {
                    "build-on": [
                        {
                            "name": "ubuntu",
                            "channel": "22.04",
                            "architectures": ["amd64", "riscv64"],
                        },
                        {
                            "name": "ubuntu",
                            "channel": "20.04",
                            "architectures": ["amd64", "arm64"],
                        },
                    ],
                    "run-on": [
                        {
                            "name": "ubuntu",
                            "channel": "22.04",
                            "architectures": ["amd64"],
                        },
                        {
                            "name": "ubuntu",
                            "channel": "22.04",
                            "architectures": ["riscv64"],
                        },
                        {
                            "name": "ubuntu",
                            "channel": "22.04",
                            "architectures": ["arm64"],
                        },
                    ],
                },
                {
                    "build-on": [{"name": "ubuntu", "channel": "20.04"}],
                    "run-on": [
                        {
                            "name": "ubuntu",
                            "channel": "20.04",
                            "architectures": [
                                "amd64",
                                "arm64",
                                "riscv64",
                                "s390x",
                                "ppc64el",
                                "armhf",
                            ],
                        }
                    ],
                },
            ],
            [
                craft_platforms.BuildInfo(
                    platform="ubuntu-22.04-amd64",
                    build_on=craft_platforms.DebianArchitecture.AMD64,
                    build_for=craft_platforms.DebianArchitecture.AMD64,
                    build_base=craft_platforms.DistroBase(
                        distribution="ubuntu",
                        series="22.04",
                    ),
                ),
                craft_platforms.BuildInfo(
                    platform="ubuntu-22.04-amd64",
                    build_on=craft_platforms.DebianArchitecture.RISCV64,
                    build_for=craft_platforms.DebianArchitecture.AMD64,
                    build_base=craft_platforms.DistroBase(
                        distribution="ubuntu",
                        series="22.04",
                    ),
                ),
                craft_platforms.BuildInfo(
                    platform="ubuntu-22.04-riscv64",
                    build_on=craft_platforms.DebianArchitecture.AMD64,
                    build_for=craft_platforms.DebianArchitecture.RISCV64,
                    build_base=craft_platforms.DistroBase(
                        distribution="ubuntu",
                        series="22.04",
                    ),
                ),
                craft_platforms.BuildInfo(
                    platform="ubuntu-22.04-riscv64",
                    build_on=craft_platforms.DebianArchitecture.RISCV64,
                    build_for=craft_platforms.DebianArchitecture.RISCV64,
                    build_base=craft_platforms.DistroBase(
                        distribution="ubuntu",
                        series="22.04",
                    ),
                ),
                craft_platforms.BuildInfo(
                    platform="ubuntu-22.04-arm64",
                    build_on=craft_platforms.DebianArchitecture.AMD64,
                    build_for=craft_platforms.DebianArchitecture.ARM64,
                    build_base=craft_platforms.DistroBase(
                        distribution="ubuntu",
                        series="22.04",
                    ),
                ),
                craft_platforms.BuildInfo(
                    platform="ubuntu-22.04-arm64",
                    build_on=craft_platforms.DebianArchitecture.RISCV64,
                    build_for=craft_platforms.DebianArchitecture.ARM64,
                    build_base=craft_platforms.DistroBase(
                        distribution="ubuntu",
                        series="22.04",
                    ),
                ),
                craft_platforms.BuildInfo(
                    platform="ubuntu-20.04-amd64",
                    build_on=craft_platforms.DebianArchitecture.AMD64,
                    build_for=craft_platforms.DebianArchitecture.AMD64,
                    build_base=craft_platforms.DistroBase(
                        distribution="ubuntu",
                        series="20.04",
                    ),
                ),
                craft_platforms.BuildInfo(
                    platform="ubuntu-20.04-amd64",
                    build_on=craft_platforms.DebianArchitecture.ARM64,
                    build_for=craft_platforms.DebianArchitecture.AMD64,
                    build_base=craft_platforms.DistroBase(
                        distribution="ubuntu",
                        series="20.04",
                    ),
                ),
                craft_platforms.BuildInfo(
                    platform="ubuntu-20.04-riscv64",
                    build_on=craft_platforms.DebianArchitecture.AMD64,
                    build_for=craft_platforms.DebianArchitecture.RISCV64,
                    build_base=craft_platforms.DistroBase(
                        distribution="ubuntu",
                        series="20.04",
                    ),
                ),
                craft_platforms.BuildInfo(
                    platform="ubuntu-20.04-riscv64",
                    build_on=craft_platforms.DebianArchitecture.ARM64,
                    build_for=craft_platforms.DebianArchitecture.RISCV64,
                    build_base=craft_platforms.DistroBase(
                        distribution="ubuntu",
                        series="20.04",
                    ),
                ),
                craft_platforms.BuildInfo(
                    platform="ubuntu-20.04-arm64",
                    build_on=craft_platforms.DebianArchitecture.AMD64,
                    build_for=craft_platforms.DebianArchitecture.ARM64,
                    build_base=craft_platforms.DistroBase(
                        distribution="ubuntu",
                        series="20.04",
                    ),
                ),
                craft_platforms.BuildInfo(
                    platform="ubuntu-20.04-arm64",
                    build_on=craft_platforms.DebianArchitecture.ARM64,
                    build_for=craft_platforms.DebianArchitecture.ARM64,
                    build_base=craft_platforms.DistroBase(
                        distribution="ubuntu",
                        series="20.04",
                    ),
                ),
                craft_platforms.BuildInfo(
                    platform="ubuntu-20.04-amd64-arm64-riscv64-s390x-ppc64el-armhf",
                    build_on=craft_platforms.DebianArchitecture.AMD64,
                    build_for=craft_platforms.DebianArchitecture.AMD64,
                    build_base=craft_platforms.DistroBase(
                        distribution="ubuntu",
                        series="20.04",
                    ),
                ),
                craft_platforms.BuildInfo(
                    platform="ubuntu-20.04-amd64-arm64-riscv64-s390x-ppc64el-armhf",
                    build_on=craft_platforms.DebianArchitecture.ARM64,
                    build_for=craft_platforms.DebianArchitecture.AMD64,
                    build_base=craft_platforms.DistroBase(
                        distribution="ubuntu",
                        series="20.04",
                    ),
                ),
                craft_platforms.BuildInfo(
                    platform="ubuntu-20.04-amd64-arm64-riscv64-s390x-ppc64el-armhf",
                    build_on=craft_platforms.DebianArchitecture.PPC64EL,
                    build_for=craft_platforms.DebianArchitecture.AMD64,
                    build_base=craft_platforms.DistroBase(
                        distribution="ubuntu",
                        series="20.04",
                    ),
                ),
                craft_platforms.BuildInfo(
                    platform="ubuntu-20.04-amd64-arm64-riscv64-s390x-ppc64el-armhf",
                    build_on=craft_platforms.DebianArchitecture.RISCV64,
                    build_for=craft_platforms.DebianArchitecture.AMD64,
                    build_base=craft_platforms.DistroBase(
                        distribution="ubuntu",
                        series="20.04",
                    ),
                ),
                craft_platforms.BuildInfo(
                    platform="ubuntu-20.04-amd64-arm64-riscv64-s390x-ppc64el-armhf",
                    build_on=craft_platforms.DebianArchitecture.S390X,
                    build_for=craft_platforms.DebianArchitecture.AMD64,
                    build_base=craft_platforms.DistroBase(
                        distribution="ubuntu",
                        series="20.04",
                    ),
                ),
            ],
        ),
    ],
)
def test_bases_build_plan_success(bases, expected):
    assert charm.get_bases_charm_build_plan(bases) == expected



# Copyright 2024 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Build info."""

import dataclasses
from typing import Literal, Union

from craft_platforms import _architectures, _distro


@dataclasses.dataclass
class BuildInfo:
    """Platform build information."""

    platform: str
    """The platform name."""

    build_on: _architectures.DebianArchitecture
    """The architecture to build on."""

    build_for: Union[_architectures.DebianArchitecture, Literal["all"]]
    """The architecture to build for."""

    build_base: _distro.DistroBase
    """The base to build on."""



# This file is part of craft-platforms.
#
# Copyright 2025 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for craft-platforms."""

from typing import Sequence, Union


def vectorize(item: Union[str, Sequence[str]]) -> Sequence[str]:
    """Convert a string into a sequence of strings.

    If the item is already a sequence, it is returned as-is.

    :param item: The item to vectorize.

    :returns: A sequence of strings.
    """
    if isinstance(item, str):
        return [item]
    return item



# This file is part of craft-platforms.
#
# Copyright 2024 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Architecture related utilities."""

from __future__ import annotations

import enum
import platform
from typing import Literal, Tuple, Union

from typing_extensions import Self

from craft_platforms import _distro


class DebianArchitecture(str, enum.Enum):
    """A Debian architecture."""

    AMD64 = "amd64"
    ARM64 = "arm64"
    ARMHF = "armhf"
    I386 = "i386"
    PPC64EL = "ppc64el"
    RISCV64 = "riscv64"
    S390X = "s390x"

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        """Generate the repr of the string value.

        This is different from the Python/StringEnum default because of the very common
        idiom in Craft codebases of using a string's repr to pretty-print to users.
        """
        return f"{self.value!r}"

    @classmethod
    def from_machine(cls, arch: str) -> Self:
        """Get a DebianArchitecture value from the given platform arch.

        :param arch: a string containing an architecture as returned by platform.machine()
        :returns: The DebianArchitecture enum value
        :raises: ValueError if the architecture is not a valid Debian architecture.
        """
        return cls(_ARCH_TRANSLATIONS_PLATFORM_TO_DEB.get(arch.lower(), arch.lower()))

    @classmethod
    def from_host(cls) -> Self:
        """Get the DebianArchitecture of the running host."""
        return cls.from_machine(platform.machine())

    def to_platform_arch(self) -> str:
        """Convert this DebianArchitecture to a platform string.

        :returns: A string matching what platform.machine() or uname -m would return.
        """
        return _ARCH_TRANSLATIONS_DEB_TO_PLATFORM.get(self.value, self.value)


# architecture translations from the platform syntax to the deb/snap syntax
_ARCH_TRANSLATIONS_PLATFORM_TO_DEB = {
    "aarch64": "arm64",
    "armv7l": "armhf",
    "i686": "i386",
    "ppc": "powerpc",
    "ppc64le": "ppc64el",
    "x86_64": "amd64",
}

# architecture translations from the deb/snap syntax to the platform syntax
_ARCH_TRANSLATIONS_DEB_TO_PLATFORM = {
    deb: platform for platform, deb in _ARCH_TRANSLATIONS_PLATFORM_TO_DEB.items()
}


def parse_base_and_architecture(
    arch: str,
) -> Tuple[_distro.DistroBase | None, Union[DebianArchitecture, Literal["all"]]]:
    """Get the debian arch and optional base from an architecture entry.

    The architecture may have an optional base prefixed as '[<base>:]<arch>'.

    :param arch: The architecture entry.

    :returns: A tuple of the DistroBase and the architecture. The architecture is either
     a DebianArchitecture or 'all'.

    :raises ValueError: If the architecture or base is invalid.
    """
    if ":" in arch:
        base_str, _, arch_str = arch.partition(":")
        base = _distro.DistroBase.from_str(base_str)
    else:
        base = None
        arch_str = arch

    try:
        return base, DebianArchitecture(arch_str) if arch_str != "all" else "all"
    except ValueError:
        raise ValueError(f"{arch_str!r} is not a valid Debian architecture.") from None



# This file is part of craft-platforms.
#
# Copyright 2025 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""General build planner for any app."""

from typing import Any, Callable, Dict, Iterable

from craft_platforms import charm, rock, snap
from craft_platforms._buildinfo import BuildInfo
from craft_platforms._platforms import get_platforms_build_plan

_APP_SPECIFIC_PLANNERS: Dict[str, Callable[..., Iterable[BuildInfo]]] = {
    "charmcraft": charm.get_charm_build_plan,
    "rockcraft": rock.get_rock_build_plan,
    "snapcraft": snap.get_platforms_snap_build_plan,
}


def get_build_plan(
    app: str,
    *,
    project_data: Dict[str, Any],
) -> Iterable[BuildInfo]:
    """Get a build plan for a given application.

    :param app: The name of the application (e.g. snapcraft, charmcraft, rockcraft)
    :param project_data: The raw dictionary of the project's YAML file. Normally this
        is what's output from ``yaml.safe_load()``.
    :returns: An iterable containing each possible BuildInfo for this file.

    This function is an abstraction layer over the general build planners, taking
    the application's name and its raw data and returning an exhaustive build plan
    for the given project for that application. This allows craft-platforms to be used
    in a forward-compatible manner as it adds more logic around selecting build
    planners or special behaviour for more apps.
    """
    planner = _APP_SPECIFIC_PLANNERS.get(app, get_platforms_build_plan)
    if app == "charmcraft":
        return planner(project_data)

    args = {
        "base": project_data.get("base"),
        "platforms": project_data.get("platforms"),
        "build_base": project_data.get("build-base"),
    }

    if app == "snapcraft":
        args["snap_type"] = project_data.get("type")

    return planner(**args)
